/*
Package:  moab
Module:   MComm.c
Version:  4.0.4
Modified: May 28, 2004

Copyright:

  Copyright (C) 1999-2004 Cluster Resources, Inc

  All Rights Reserved

Disclaimer:

 THIS SOFTWARE IS PROVIDED AS IS AND CLUSTER RESOURCES INC (CRI) DISCLAIMS ALL WARRANTIES RELATING TO THE SOFTWARE, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  NEITHER CRI, NOR ANYONE INVOLVED IN THE CREATION, PRODUCTION, OR DELIVERY OF THE SOFTWARE SHALL BE LIABLE FOR ANY INDIRECT, CONSEQUENTIAL, OR INCIDENTAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE SOFTWARE EVEN IF CRI HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES OR CLAIMS.  THE PERSON OR ENTITY USING THE SOFTWARE BEARS ALL RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE.
*/

#include "msu-internal.h"

extern const char *SUMKType[];

void SUCommParserInit(

  sucomm_parser_t *parser) 

  {
  parser->length = 0;
  parser->state = 0;
  parser->msgSize = 0;
  
  return;
  }



int SUCommParse(

  su_t            *su,
  sucomm_parser_t *parser,
  void            *data,
  size_t           size,
  int             *done)

  {
  *done = 0;

  if (parser->length + size >= MAX_SUBUFFER) /* buffer overflow */
    {
    return(FAILURE);
    }

  memcpy(parser->buffer + parser->length, data, size);
  parser->length += size;

  switch (parser->state)
    {
    case 0: /* initial */
        if (parser->length < 9)
          return(SUCCESS);

        parser->buffer[8] = 0;
        parser->msgSize = strtoul(parser->buffer, NULL, 10);

        if (parser->msgSize == 0 || parser->msgSize >= MAX_SUBUFFER)
          return(FAILURE);

        memmove(parser->buffer, parser->buffer + 9, parser->length - 9);
		parser->length -= 9;
        parser->state = 1;
        /* fall through */
    case 1: /* length known */
        if (parser->length >= parser->msgSize) {
          *done = 1;
          parser->buffer[parser->length] = 0;
        }

    return(SUCCESS);
    }

  return(FAILURE);
  }  /* END SUCommParse() */




int SUCommWrapMessage(

  su_t       *su,
  const char *cmd,
  const char *args,
  suclcred_t *Cred,
  char       *buffer,
  size_t      size)

  {

  char msg[MAX_SUBUFFER];
  char cksum[MAX_SULINE];
  time_t Now;
  char backup;

  time(&Now);

  snprintf(msg, MAX_SUBUFFER, "%s%ld ID=%s %s%s%s %s%s %s%s %s%s",
    SUMKType[sumkTimeStamp],
    (long)Now,
    su->Env->UserName,
    SUMKType[sumkData],
    SUMKType[sumkCommand],
    cmd,
    SUMKType[sumkAuth],
    (su->Env->UserName != NULL) ? su->Env->UserName : NONE,
    SUMKType[sumkClient],
    (su->Name != NULL) ? su->Name : NONE,
    SUMKType[sumkArgs],
    (args != NULL) ? args : "");

  SUSecGetMChecksum2(
    su,
    msg,
    strlen(msg),
    NULL,
    0,
    Cred->EncryptionData.CSKey,
    cksum);

  snprintf(buffer + 9, size - 9, "%s%s %s",
    SUMKType[sumkCheckSum],
    cksum,
    msg);

  backup = buffer[9];
  sprintf(buffer, "%08ld\n", (long)strlen(buffer+9));
  buffer[9] = backup;

  return strlen(buffer);
  }  /* END SUCommWrapMessage() */




int SUCommMessageIsValid(

  su_t       *su,
  char       *msg, 
  suclcred_t *Cred)

  {
  char  *cksum = strstr(msg,SUMKType[sumkCheckSum]);
  char  *ts = strstr(msg,SUMKType[sumkTimeStamp]);
  char  *data = strstr(msg,SUMKType[sumkData]);
  char  *cksumEnd;
  char  *tsEnd;
  char   CKSum[MAX_SULINE];
  time_t TSVal;
  time_t Now;

  if (cksum == NULL)
    {
    DBG(su,1,sfSocket) SUEMLog(su,"ALERT:    cannot locate checksum '%s'\n",
      msg);

    return(FAILURE);
    }

  if (ts == NULL)
    {
    DBG(su,1,sfSocket) SUEMLog(su,"ALERT:    cannot locate timestamp '%s'\n",
      msg);

    return(FAILURE);
    }

  if (data == NULL)
    {
    DBG(su,1,sfSocket) SUEMLog(su,"ALERT:    cannot locate data '%s'\n",
      msg);

    return(FAILURE);
    }

  cksum += strlen(SUMKType[sumkCheckSum]);
  cksumEnd = strchr(cksum,' ');
  if (cksumEnd)
      *cksumEnd = 0;

  SUSecGetMChecksum(
      su,
      ts,
      strlen(ts),
      Cred->EncryptionData.CSKey,
      CKSum);

  if (strcmp(cksum,CKSum) != 0)
    {
    DBG(su,1,sfSocket) SUEMLog(su,"ALERT:    checksum does not match (%s:%s)\n",
      cksum,
      CKSum);

    if (cksumEnd)
      *cksumEnd = ' ';

    return(FAILURE);
    }

  if (cksumEnd)
    *cksumEnd = ' ';

  ts += strlen(SUMKType[sumkTimeStamp]);
  tsEnd = strchr(ts, ' ');
  if (tsEnd)
      *tsEnd = 0;

  TSVal = strtoul(ts, NULL, 10);

  if (tsEnd)
      *tsEnd = ' ';

  time(&Now);

  if (((Now - TSVal) > 3600) || ((Now - TSVal) < -3600))
    {
    DBG(su,1,sfSocket) SUEMLog(su,"ALERT:    timestamp does not match (%ld:%ld)\n",
      (unsigned long)Now,
      (unsigned long)TSVal);

      return(FAILURE);
      }

  return(SUCCESS);
  }  /* END SUCommMessageIsValid() */


/* END MComm.c */

