/*
Package:  moab
Module:   MSSSI.c
Version:  4.0.4
Modified: May 28, 2004

Copyright:

  Copyright (C) 1999-2004 Cluster Resources, Inc

  All Rights Reserved

Disclaimer:

 THIS SOFTWARE IS PROVIDED AS IS AND CLUSTER RESOURCES INC (CRI) DISCLAIMS ALL WARRANTIES RELATING TO THE SOFTWARE, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  NEITHER CRI, NOR ANYONE INVOLVED IN THE CREATION, PRODUCTION, OR DELIVERY OF THE SOFTWARE SHALL BE LIABLE FOR ANY INDIRECT, CONSEQUENTIAL, OR INCIDENTAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE SOFTWARE EVEN IF CRI HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES OR CLAIMS.  THE PERSON OR ENTITY USING THE SOFTWARE BEARS ALL RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE.
*/

/* Contains:                                   *
 *  int MSSSDoCommand(P,CmdString,OBuf,OE)     *
 *                                             */

#include "moab.h"
#include "moab-proto.h"  

#include "mg2.h"

#define __MSSS30

extern mlog_t    mlog;

extern msched_t    MSched;
extern mclass_t    MClass[];
extern mjob_t     *MJob[];
extern mbool_t     MGlobalTLock;
extern mrm_t       MRM[];    
extern mattrlist_t MAList;
extern mnode_t    *MNode[];

extern const char *MJobState[];
extern const char *MNodeState[];
extern const char *MSSSJobState[];
extern const char *MSSSCName[];
extern const char *MSSSSockProtocol[];
extern const char *MClassState[];
extern const char *MSSSVName[];

extern char *MSON[];
extern char *MSAN[];

typedef struct {
  msocket_t S;
  } ssshandle_t;
 
typedef struct {
  int          SchedSD;
  int          ServerSD;
  ssshandle_t  QMH;
  ssshandle_t  NMH;
  int          UpdateIteration;
  char        *WorkloadBuffer;
  char        *ResourceBuffer;
  char        *QueueBuffer;
  } sssrm_t;

/* SSS prototypes */

int MSSSInitialize(mrm_t *,int *);
int MSSSGetData(mrm_t *,int *);
int MSSSJobStart(mjob_t *,mrm_t *,char *,int *);
int MSSSJobRequeue(mjob_t *,mrm_t *,mjob_t **,char *,int *); 
int MSSSJobModify(mjob_t *,mrm_t *,char *,char *,char *,int *);
int MSSSJobCancel(mjob_t *,mrm_t *,char *,char *,int *); 
int MSSSJobSubmit(char *,mrm_t *,mjob_t **,char *,char *,enum MStatusCodeEnum *);
int MSSSJobSuspend(mjob_t *,mrm_t *,char *,int *); 
int MSSSJobResume(mjob_t *,mrm_t *,char *,int *); 
int MSSSJobCheckpoint(mjob_t *,mrm_t *,mbool_t,char *,int *);
int MSSSClusterQuery(mrm_t *,int *,char *,enum MStatusCodeEnum *);
int MSSSWorkloadQuery(mrm_t *,int *,char *,enum MStatusCodeEnum *);
int MSSSQueueQuery(mrm_t *,int *,char *,enum MStatusCodeEnum *);
int MSSSJobLoad(char *,void *,mjob_t **,short *,mrm_t *);
int MSSSJobUpdate(void *,mjob_t **,short *,mrm_t *);
int MSSSNodeLoad(mnode_t *,void *,int,mrm_t *);
int MSSSNodeUpdate(mnode_t *,void *,int,mrm_t *);
int __MSSSJobStepGetState(mxml_t *,mrm_t *,char *,enum MJobStateEnum *); 
int __MSSSNodeGetState(void *,mrm_t *,char *,enum MNodeStateEnum *); 
int __MSSSQueueGetState(mxml_t *,mrm_t *,char *,int *);
int MSSSGetQueueInfo(mnode_t *,char C[][MMAX_NAME],char A[][MMAX_NAME]);
int __MSSSGetTaskList(char *,short *,int *,int *);
int __MSSSJobSetAttr(mjob_t *,int,char *,char *,mbool_t,void *);
int __MSSSNodeSetAttr(mnode_t *,mrm_t *,char *,char *,void *);
int __MSSSQueueSetAttr(mclass_t *,char *,char *);
char *__MSSSGetErrMsg(int);
int MSSSEncapsulateMessage(char *,char *,mbool_t,char **,long *);
int MSSSSetup(int);

/* END SSS prototypes */


/* attribute           object    0.2 attr                  2.0 attr        3.0 attr  */

const mobjattr_t MSSSJobTable[] =  { 
  { mjaAccount,        mxoJob, { "accountName",            "ProjectId",    NULL, NULL } },
  { mjaAllocNodeList,  mxoJob, { NULL,                     "NodeList",     NULL, NULL } },
  { mjaArgs,           mxoJob, { NULL,                     "Arguments",    NULL, NULL } },
  { mjaCompletionTime, mxoJob, { NULL,                     "CompletionTime", NULL, NULL } },
  { mjaExecutable,     mxoJob, { NULL,                     "Executable",   NULL, NULL } },
  { mjaGroup,          mxoJob, { "groupName",              "GroupId",      NULL, NULL } },
  { mjaIWD,            mxoJob, { NULL,                     "InitialWorkingDirectory", NULL, NULL } },
  { mjaJobName,        mxoJob, { "jobName",                "JobName",      NULL, NULL } },
  { mjaJobID,          mxoJob, { "jobId",                  "JobId",        NULL, NULL } },
  { mjaQOS,            mxoJob, { "qos",                    "QualityOfService", NULL, NULL } },
  { mjaUser,           mxoJob, { "userName",               "UserId",       NULL, NULL } },
  { mjaStartTime,      mxoJob, { NULL,                     "CompletionTime", NULL, NULL } },
  { mjaState,          mxoJob, { "jobState",               "JobState",     NULL, NULL } },
  { mjaReqNodes,       mxoJob, { "requestedNodes",         "NodeCount",    NULL, NULL } },
  { mrqaReqClass,      mxoReq, { "requestedQueue",         "Queue",        NULL, NULL } }, /* FIX */
  { mrqaTCReqMin,      mxoReq, { "requestedProcessors",    "Processors",   NULL, NULL } },
  { mjaReqAWDuration,  mxoJob, { "wallclockLimit",         "WallDuration", NULL, NULL } }, 
  { mjaAWDuration,     mxoJob, { "wallclockTimeUsed",      "WallDuration", NULL, NULL } },
  { mjaReqSMinTime,    mxoJob, { "eligibleDate",           "EligibleDate", NULL, NULL } },
  { mjaRMXString,      mxoJob, { "opaqueFlags",            "OpaqueFlags",  NULL, NULL } },
  { mjaStepID,         mxoJob, { NULL,                     "StepId",       NULL, NULL } },
  { mjaStartTime,      mxoJob, { NULL,                     "StartTime",    NULL, NULL } },
  { mjaSubmitTime,     mxoJob, { NULL,                     "SubmitTime",   NULL, NULL } },
  { -1,                -1,     { NULL,                     NULL,           NULL, NULL } }};

/* handle multi-level attribute population */

/* attribute          object    0.2 attr         2.0 attr    3.0 attr      xpath */

const mobjattr_t MSSSNodeTable[] = {
  { mnaFeatures,    mxoNode,  { "features",      "Features", NULL,         "Features" } },
  { mnaNodeID,      mxoNode,  { "nodeId",        "NodeId", NULL,           "NodeId" } },
  { mnaNodeState,   mxoNode,  { "nodeState",     "NodeState", NULL,        "NodeState" } },
  { mnaRCMem,       mxoNode,  { "memTot",        "ConfiguredMemory", NULL, "Configured/Memory" } },
  { mnaRAMem,       mxoNode,  { "memFree",       "AvailableMemory", NULL,  "Available/Memory" } },
  { mnaRCSwap,      mxoNode,  { "swapTot",       "ConfiguredSwap", NULL,   "Configured/Swap" } },
  { mnaRASwap,      mxoNode,  { "swapFree",      "AvailableSwap", NULL,    "Available/Swap" } },
  { mnaRCDisk,      mxoNode,  { "diskTot",       "ConfiguredDisk", NULL,   "Configured/Disk" } },
  { mnaRADisk,      mxoNode,  { "diskFree",      "AvailableDisk", NULL,    "Available/Disk" } },
  { mnaRCProc,      mxoNode,  { "numProcessors", "ConfiguredProcessors", NULL, "Configured/Processors" } }, 
  { mnaLoad,        mxoNode,  { "utlProcessors", "UtilizedProcessors", NULL, "Utilized/Processors" } },

/* { mna???,         mxoNode,  { "nodeType",     "NodeType", NULL, NULL } }, */
/* { mna???,         mxoNode,  { "maxtask",      "VirtualProcessors", NULL, NULL } }, */

  /* NOTE:  percentCPUUsed is available, node load is not */

  { mnaRAProc,      mxoNode,  { "AVLProcCount",  "AvailableProcessors", NULL, "Available/Processors" } },
  { -1,             -1,       { NULL,            NULL, NULL, NULL } }};

const mobjattr_t MSSSNodeRTable[] = {
  { mrDisk,         mxoNode,  { "disk",          "Disk",       NULL,     NULL } },
  { mrMem,          mxoNode,  { "memory",        "Memory",     NULL,     NULL } },
  { mrProc,         mxoNode,  { "processors",    "Processors", NULL,     NULL } },
  { mrSwap,         mxoNode,  { "swap",          "Swap",       NULL,     NULL } },
  { -1,             -1,       { NULL,            NULL, NULL, NULL } }};


const mobjattr_t MSSSQueueTable[] = {
  { mclaName,       mxoClass, { "queueId",       "QueueId", NULL, NULL } },
  { mclaState,      mxoClass, { "queueStarted",  "QueueStarted", NULL, NULL } },
  { mclaHostList,   mxoClass, { "hostList",      "HostList", NULL, NULL } },
  { -1,             -1,       { NULL,            NULL, NULL, NULL } }};

const mobjattr_t MSSSOTable[] = {
  { mxoJob,         mxoJob,   { "job",           "Job", NULL, NULL } },
  { mxoNode,        mxoNode,  { "node",          "Node", NULL, NULL } },
  { mxoQueue,       mxoQueue, { "queue",         "Queue", NULL, NULL } },
  { mxoNONE,        mxoNONE,  { NULL,            NULL, NULL, NULL } } };

const mobjattr_t MSSSJATable[] = {
  { mrmJobCancel,     mxoNONE,  { "delete",        "Cancel", NULL, NULL } },
  { mrmJobCheckpoint, mxoNONE,  { "checkpoint",    "Checkpoint", NULL, NULL } },
  { mrmJobModify,     mxoNONE,  { "modify",        "Modify", NULL, NULL } },
  { mrmJobQuery,      mxoNONE,  { "query",         "Query", NULL, NULL } },
  { mrmJobRequeue,    mxoNONE,  { "requeue",       "Requeue", NULL, NULL } },
  { mrmJobResume,     mxoNONE,  { "resume",        "Resume", NULL, NULL } },
  { mrmJobStart,      mxoNONE,  { "start",         "Start", NULL, NULL } },
  { mrmJobSubmit,     mxoNONE,  { "submit",        "Submit", NULL, NULL } },
  { mrmJobSuspend,    mxoNONE,  { "suspend",       "Suspend", NULL, NULL } },
  { mrmNONE,          mxoNONE,  { NULL,            NULL, NULL, NULL } } };

char *MSSSJobAttr[MMAX_SSSVERS][MMAX_SSSATTR];
char *MSSSReqAttr[MMAX_SSSVERS][MMAX_SSSATTR];
char *MSSSNodeAttr[MMAX_SSSVERS][MMAX_SSSATTR];
char *MSSSNodeRAttr[MMAX_SSSVERS][MMAX_SSSATTR];
char *MSSSQueueAttr[MMAX_SSSATTR];
char *MSSSObjName[MMAX_SSSVERS][MMAX_SSSATTR];
char *MSSSJAName[MMAX_SSSVERS][MMAX_SSSJACTION];


#ifdef __M32COMPAT

#define MJOBREMOVE(J)  MJobRemove(*(J))
#define MJOBPROCESSREMOVED(J)  MJobProcessRemoved(*(J))
#define MJOBPROCESSCOMPLETED(J)  MJobProcessCompleted(*(J))

#define SRM RM

#else /* __M32COMPAT */

#define MJOBREMOVE(J)  MJobRemove(J)
#define MJOBPROCESSREMOVED(J)  MJobProcessRemoved(J)
#define MJOBPROCESSCOMPLETED(J)  MJobProcessCompleted(J)

#endif /* __M32COMPAT */



int MSSSSetup(

  int VIndex)  /* I */

  {
  int aindex;
  int vindex;

  static const char *ANone = "[NONE]";

  static mbool_t IsInitialized = FALSE;

  if (IsInitialized == TRUE)
    {
    return(SUCCESS);
    }

  /* NOTE:  populate table with '[NONE]' */

  for (vindex = 0;vindex < MMAX_SSSVERS;vindex++)
    {
    for (aindex = 0;aindex < MMAX_SSSATTR;aindex++)
      {
      MSSSJobAttr[vindex][aindex] = (char *)ANone;
      MSSSReqAttr[vindex][aindex] = (char *)ANone;

      MSSSObjName[vindex][aindex] = (char *)ANone;
      MSSSNodeAttr[vindex][aindex] = (char *)ANone;
      MSSSNodeRAttr[vindex][aindex] = (char *)ANone;

      if (vindex > 0)
        continue;
 
      MSSSQueueAttr[aindex] = (char *)ANone;
      }  /* END for (aindex) */

    for (aindex = 0;MSSSJobTable[aindex].AIndex != -1;aindex++)
      {
      if (MSSSJobTable[aindex].OType == mxoJob)
        {
        MSSSJobAttr[vindex][MSSSJobTable[aindex].AIndex] =
          MSSSJobTable[aindex].AName[vindex];
        }
      else
        {
        MSSSReqAttr[vindex][MSSSJobTable[aindex].AIndex] =
          MSSSJobTable[aindex].AName[vindex];
        }
      }  /* END for (aindex) */

    for (aindex = 0;MSSSOTable[aindex].AIndex != mxoNONE;aindex++)
      {
      MSSSObjName[vindex][MSSSOTable[aindex].AIndex] =
        MSSSOTable[aindex].AName[vindex];
      }  /* END for (aindex) */

    for (aindex = 0;MSSSJATable[aindex].AIndex != mrmNONE;aindex++)
      {
      MSSSJAName[vindex][MSSSJATable[aindex].AIndex] =
        MSSSJATable[aindex].AName[vindex];
      }  /* END for (aindex) */

    for (aindex = 0;MSSSNodeRTable[aindex].AIndex != -1;aindex++)
      {
      if (MSSSNodeRTable[aindex].OType == mxoNode)
        {
        MSSSNodeRAttr[vindex][MSSSNodeRTable[aindex].AIndex] =
          MSSSNodeRTable[aindex].AName[vindex];
        }
      else
        {
        /* NO-OP */
        }
      }  /* END for (aindex) */

    for (aindex = 0;MSSSNodeTable[aindex].AIndex != -1;aindex++)
      {
      if (MSSSNodeTable[aindex].OType == mxoNode)
        {
        MSSSNodeAttr[vindex][MSSSNodeTable[aindex].AIndex] =
          MSSSNodeTable[aindex].AName[vindex];
        }
      else
        {
        /* NO-OP */
        }
      }  /* END for (aindex) */
    }    /* END for (vindex) */

  for (aindex = 0;MSSSQueueTable[aindex].AIndex != -1;aindex++)
    {
    if (MSSSQueueTable[aindex].OType == mxoClass)
      {
      MSSSQueueAttr[MSSSQueueTable[aindex].AIndex] =
        MSSSQueueTable[aindex].AName[1];  /* NOTE:  '1' is temp hack */
      }
    else
      {
      /* NO-OP */
      }
    }  /* END for (aindex) */

  IsInitialized = TRUE;

  return(SUCCESS);
  }    /* END MSSSSetup() */





int MSSSLoadModule(
 
  mrmfunc_t *F)  /* I */
 
  {
  if (F == NULL)
    {
    return(FAILURE);
    }
 
  F->ClusterQuery   = MSSSClusterQuery;
  F->JobCancel      = MSSSJobCancel;
  F->JobCheckpoint  = MSSSJobCheckpoint;
  F->JobModify      = NULL;
  F->JobQuery       = NULL;
  F->JobRequeue     = MSSSJobRequeue;
  F->JobResume      = MSSSJobResume;
  F->JobStart       = MSSSJobStart;
  F->JobSubmit      = MSSSJobSubmit;
  F->JobSuspend     = MSSSJobSuspend;
  F->QueueQuery     = MSSSQueueQuery;
  F->ResourceModify = NULL;
  F->ResourceQuery  = NULL;
  F->RMInitialize   = MSSSInitialize;
  F->RMQuery        = NULL;
  F->WorkloadQuery  = MSSSWorkloadQuery;
 
  return(SUCCESS);
  }  /* END MSSSLoadModule() */




int MSSSInitialize(

  mrm_t *R,   /* I */
  int   *SC)  /* I */
 
  {
  sssrm_t *S;

  const char *FName = "MSSSInitialize";

  MDB(1,fSSS) MLog("%s(%s,SC)\n",
    FName,
    (R != NULL) ? R->Name : "NULL");
 
  if (R == NULL)
    {
    return(FAILURE);
    }
 
  if (R->S == NULL)
    {
    R->S = (void *)calloc(1,sizeof(sssrm_t));
 
    /* set default rm specific values */
 
    /* NYI */
    }

  R->P[0].Type = mpstQM;
  R->P[0].S    = &((sssrm_t *)R->S)->QMH.S;

  R->P[1].Type = mpstNM;
  R->P[1].S    = &((sssrm_t *)R->S)->NMH.S;

  ((sssrm_t *)(R->S))->UpdateIteration = -1;

  /* set SSS_ROOT */

  MUSetEnv("SSS_ROOT","/usr/local/sss");

  S = (sssrm_t *)R->S;

  R->Version = MUGetIndex(R->APIVersion,MSSSVName,FALSE,msssV0_2);

  MSSSSetup(R->Version);
 
  /* register with directory service */

  /* NOTE:  should loop through all supported protocols (NYI) */

  if (MSysDSRegister(
       (char *)MSSSCName[mpstSC],
       MRM[0].Name,
       MSched.ServerHost,
       MSched.ServerPort,
       NULL,
       (char *)MSSSSockProtocol[mspSingleUseTCP]) == FAILURE)
    {
    MDB(1,fRM) MLog("ALERT:    cannot register with directory service\n");
    }

  if (R->UseDirectoryService == TRUE)
    { 
    char HostName[MMAX_NAME];
    int  Port;
    char WireProtocol[MMAX_NAME];
    char SocketProtocol[MMAX_NAME];

    /* obtain interface information */

    if (MSysDSQuery(
         (char *)MSSSCName[mpstQM],
         MRM[0].Name,
         HostName,
         &Port,
         WireProtocol,
         SocketProtocol) == SUCCESS)
      {
      MDB(1,fRM) MLog("INFO:     obtained QM info (H: %s  P: %d  S: %s)\n",
        HostName,
        Port,
        SocketProtocol);

      /* update active QM interface */

      /* NYI */
      }

    if (MSysDSQuery(
         (char *)MSSSCName[mpstNM],
         MRM[0].Name,
         HostName,
         &Port,
         WireProtocol,
         SocketProtocol) == SUCCESS)
      {
      /* update active SM interface */

      /* NYI */
      }
    }    /* END if (R->UseDirectoryService == TRUE) */

  {
  /* enable EM service */

  char HostName[MMAX_NAME];
  int  Port;
  char WireProtocol[MMAX_NAME];
  char SocketProtocol[MMAX_NAME];

  if (MSysDSQuery(
       (char *)MSSSCName[mpstEM],
       MRM[0].Name,
       HostName,
       &Port,
       WireProtocol,
       SocketProtocol) == SUCCESS)
    {
    /* update active EM interface */

    if (HostName[0] != '\0')
      MUStrDup(&MSched.EM.HostName,HostName);

    if (Port != -1)
      MSched.EM.Port = Port;
    }

  MSysEMSubmit(
    &MSched.EM,
    (char *)MSSSCName[mpstSC],
    "schedstart",
    MSched.Name);

  /* NOTE:  currently ignore EM responses */

  MSysEMRegister(
    &MSched.EM,
    (char *)MSSSCName[mpstQM],
    (char *)"job-submitted",
    NULL,
    (char *)"NONE");

  MSysEMRegister(
    &MSched.EM,
    (char *)MSSSCName[mpstQM],
    (char *)"job-finished",
    NULL,
    (char *)"NONE");
  }    /* END BLOCK */

  MDB(1,fRM) MLog("INFO:     connected to SSS server\n");

  return(SUCCESS);
  }  /* END MSSSInitialize() */




int MSSSProcessEvent(

  mrm_t *R)  /* I */

  {
  int  Cmd;
  int *Cptr;

  int EventReceived     = FALSE;
  int EventsOutstanding = TRUE;

  struct timeval timeout;    
 
  sssrm_t *S;

  msocket_t tmpS;
  msocket_t tmpC;

  if (R == NULL)
    {
    return(FAILURE);
    }

  S = (sssrm_t *)R->S;
 
  if (S->SchedSD == -1)
    {
    MDB(1,fSSS) MLog("INFO:     invalid SSS sched socket\n");
   
    return(FAILURE);
    }

  while (EventsOutstanding == TRUE)
    {       
    fd_set fdset;
    int rpp_fd = 0;   

    FD_ZERO(&fdset);

    if (rpp_fd != -1)
      {
      FD_SET(rpp_fd,&fdset);
      }

    FD_SET(S->SchedSD,&fdset);

    timeout.tv_sec  = 0;
    timeout.tv_usec = 10000;   

    if (select(FD_SETSIZE,&fdset,NULL,NULL,&timeout) == -1) 
      {
      MDB(1,fSSS) MLog("ALERT:    select failed checking SSS sched socket\n");     

      if (errno != EINTR) 
        {
        /* bad failure */

        EventsOutstanding = FALSE;
 
        break;             
        }

      EventsOutstanding = FALSE;
 
      break;     
      }
 
    if ((rpp_fd != -1) && !FD_ISSET(rpp_fd,&fdset)) 
      {
      /* rpp failure */

      MDB(9,fSSS) MLog("ALERT:    no SSS RPP sched socket connections ready\n");
 
      EventsOutstanding = FALSE;        
      }
    else if (!FD_ISSET(S->SchedSD,&fdset))
      {
      /* no connections ready */

      MDB(9,fSSS) MLog("INFO:     no SSS sched socket connections ready\n");    

      EventsOutstanding = FALSE;
 
      break;     
      }

    memset(&tmpS,0,sizeof(tmpS));
    tmpS.sd = S->SchedSD;

    memset(&tmpC,0,sizeof(tmpC));         

    if (MSUAcceptClient(&tmpS,&tmpC,NULL,(1 << msftTCP)) == FAILURE)
      {
      /* cannot accept new socket */

      MDB(1,fSSS) MLog("ALERT:    cannot accept client on SSS sched socket\n");  

      EventsOutstanding = FALSE;
 
      break;     
      }

    Cptr = &Cmd;

    if (MSURecvPacket(
          tmpS.sd,
          (char **)&Cptr,
          sizeof(int),
	  NULL,
          MDEF_CLIENTTIMEOUT) == FAILURE)
      {
      MDB(1,fSSS) MLog("ALERT:    cannot read on SSS sched socket\n");  

      close(tmpS.sd);

      EventsOutstanding = FALSE;
 
      break;      
      }

    /* SSS sched command received */

    close(tmpS.sd);

    MDB(4,fSSS) MLog("INFO:     SSS command %d received\n",
      Cmd);

    EventReceived = TRUE;         
    }  /* END while (EventsOutstanding == TRUE) */

  if (EventReceived == FALSE)
    {
    return(FAILURE);
    }
 
  return(SUCCESS);
  }  /* END MSSSProcessEvent() */ 




int MSSSGetData(

  mrm_t *R,  /* I */
  int   *SC) /* O (optional) */

  {
  int      rc;

  sssrm_t *S;

  mpsi_t  *P;

  mxml_t  *E = NULL;
  mxml_t  *RE = NULL;

  mxml_t  *C = NULL;

  mxml_t  *RDE;

  static char *JobReqString = NULL;
  static char *ResReqString = NULL;
  static char *QueueReqString = NULL;

  char tmpString[MMAX_LINE];

  const char *FName = "MSSSGetData";

  MDB(1,fSSS) MLog("%s(%s,SC)\n",
    FName,
    (R != NULL) ? R->Name : "NULL");

  if ((R == NULL) || (R->S == NULL))
    {
    return(FAILURE);
    }

  S = (sssrm_t *)R->S;      

  if (S->UpdateIteration == MSched.Iteration)
    {
    /* data already updated */

    MDB(1,fSSS) MLog("INFO:     SSS data already up to date\n");       

    return(SUCCESS);
    }

  /* free old data */

  switch(R->Version)
    {
    case msssV2_0:

      MXMLDestroyE((mxml_t **)&S->WorkloadBuffer);
      MXMLDestroyE((mxml_t **)&S->ResourceBuffer);
      MXMLDestroyE((mxml_t **)&S->QueueBuffer);

      break;

    case msssV0_2:
    default:

      MUFree(&S->WorkloadBuffer);
      MUFree(&S->ResourceBuffer);
      MUFree(&S->QueueBuffer);

      break;
    }  /* END switch(R->Version) */

  /* form SSS XML full resource query */

  if (QueueReqString == NULL)
    {
    /* build  monitor resource request */

    /* form SSS XML full workload query (NM) */

    if (R->Version == msssV2_0)
      {
      MXMLCreateE(&E,MSON[msonRequest]);

      MXMLSetAttr(E,MSAN[msanAction],(void *)MSSSJAName[R->Version][mrmJobQuery],mdfString);

      MSSSSetObject(E,MSSSObjName[R->Version][mxoQueue],NULL);
      }

    MXMLToString(E,tmpString,sizeof(tmpString),NULL,TRUE);

    MXMLDestroyE(&E);

    QueueReqString = strdup(tmpString);

    /* NOTE:  only query once at start up (temp) */

    MSSSDoCommand(
      &R->P[0],
      QueueReqString,
      NULL,
      &RDE);

    if (R->Version == msssV2_0)
      {
      MMOVEPTR((char *)RDE,S->QueueBuffer);

      MDB(1,fSSS) MLog("INFO:     empty data element read on queue query\n");

      MSUFree(R->P[0].S);
      }
    }    /* END if (QueueReqString == NULL) */

  if (JobReqString == NULL)
    {
    /* form SSS XML full workload query */

    /* NOTE:  request is dependent upon language version */

    if (R->Version == msssV2_0)
      {
      int aindex;

      const enum MJobAttrEnum JAttrList[] = {
        mjaJobName,
        mjaJobID,
        mjaUser,
        mjaGroup,
        mjaAccount,
        mjaState,
        mjaReqNodes,
        mjaAWDuration,
        mjaReqSMinTime,
        mjaRMXString,
        mjaAllocNodeList,
        mjaSubmitTime,
        mjaStartTime,
        mjaCompletionTime,
        mjaNONE };

      const enum MReqAttrEnum RAttrList[] = {
        mrqaReqClass,
        mrqaTCReqMin,
        mrqaNONE };

      MXMLCreateE(&E,MSON[msonRequest]);

      MXMLSetAttr(E,MSAN[msanAction],(void *)"Query",mdfString);
      MSSSSetObject(E,MSSSObjName[R->Version][mxoJob],NULL);

      /* specify query constraints */

      MSSSAddWhere(E,"JobClass","active");

      /* request attributes */

      for (aindex = 0;JAttrList[aindex] != mjaNONE;aindex++)
        {
        MSSSAddGet(E,MSSSJobAttr[R->Version][JAttrList[aindex]],NULL);
        }  /* END for (aindex) */

      for (aindex = 0;RAttrList[aindex] != mrqaNONE;aindex++)
        {
        MSSSAddGet(E,MSSSReqAttr[R->Version][RAttrList[aindex]],NULL);
        }  /* END for (aindex) */
      }

    MXMLToString(E,tmpString,sizeof(tmpString),NULL,TRUE);

    MXMLDestroyE(&E);   

    JobReqString = strdup(tmpString);
    }    /* END if (JobReqString == NULL) */

  MSSSDoCommand(
    &R->P[0],
    JobReqString,
    NULL,
    &RDE);

  if (R->Version == msssV2_0)
    {
    MMOVEPTR((char *)RDE,S->WorkloadBuffer);

      MDB(1,fSSS) MLog("INFO:     empty data element read on workload query\n");

    MSUFree(R->P[0].S);
    }
  else
    {
    MMOVEPTR(R->P[0].S->RBuffer,S->WorkloadBuffer);
    }

  /* form SSS XML full resource query */

  if (ResReqString == NULL)
    {
    /* build node monitor resource request */

    /* form SSS XML full resource query (NM) */

    if (R->Version == msssV2_0)
      {
      int     aindex;

      enum MNodeAttrEnum NAttrList[] = {
        mnaNodeID,
        mnaNodeState,
        mnaRCProc,
        mnaRCMem,
        mnaRAMem,
        mnaRCSwap,
        mnaRASwap,
        mnaNONE };

      MXMLCreateE(&E,MSON[msonRequest]);

      MXMLSetAttr(E,MSAN[msanAction],(void *)"Query",mdfString);

      MSSSSetObject(E,MSSSObjName[R->Version][mxoNode],NULL);

      for (aindex = 0;NAttrList[aindex] != mnaNONE;aindex++)
        {
        MSSSAddGet(E,MSSSNodeAttr[MCONST_SSSXPATH][NAttrList[aindex]],NULL);
        }  /* END for (aindex) */
      }
 
    MXMLToString(E,tmpString,sizeof(tmpString),NULL,TRUE);
 
    MXMLDestroyE(&E);
 
    ResReqString = strdup(tmpString);
    }  /* END if (ResReqString == NULL) */

  P = (R->P[1].Port == 0) ? &R->P[0] : &R->P[1];

  rc = MSSSDoCommand(
    P,
    ResReqString,
    NULL,
    &RDE);

  if (R->Version == msssV2_0)
    {
    MMOVEPTR((char *)RDE,S->ResourceBuffer);

    MDB(1,fSSS) MLog("INFO:     empty data element read on resource query\n");

    MSUFree(P->S);
    }
  else
    {
    MMOVEPTR(P->S->RBuffer,S->ResourceBuffer);
    }

  S->UpdateIteration = MSched.Iteration;     

  MDB(1,fSSS) MLog("INFO:     SSS data updated for iteration %d (rc: %d)\n",
    S->UpdateIteration,
    rc);       

  return(SUCCESS);
  }  /* END MSSSGetData() */



 
int MSSSQueueQuery(

  mrm_t                *R,      /* I */
  int                  *QCount, /* O */
  char                 *EMsg,   /* O (optional) */
  enum MStatusCodeEnum *SC)     /* O (optional) */

  {
  mclass_t *C;

  char   QID[MMAX_NAME];
  int    Status;

  char  *ptr;

  sssrm_t *S;

  mxml_t *E = NULL;
  mxml_t *RE;
  mxml_t *DE = NULL;
  mxml_t *QE;

  mnode_t *N;

  mnalloc_t *NL;

  int      CTok;

  int      cindex;
  int      nindex;

  const char *FName = "MSSSQueueQuery";

  MDB(1,fSSS) MLog("%s(%s,QCount,EMsg,SC)\n",
    FName,
    (R != NULL) ? R->Name : "NULL");

  if (QCount != NULL)
    *QCount = 0;

  if ((R == NULL) || (R->S == NULL))
    {
    return(FAILURE);
    }

  S = (sssrm_t *)R->S;

  if (S->UpdateIteration != MSched.Iteration)
    {
    MSSSGetData(R,NULL);
    }

  /* step through list */

  if (S->QueueBuffer == NULL)
    {
    return(FAILURE);
    }

  if (R->Version == msssV2_0)
    {
    DE = (mxml_t *)S->QueueBuffer;
    }
  else if (R->Version == msssV0_2)
    {
    if (((ptr = strstr(S->QueueBuffer,"<Body")) == NULL) ||
         (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
      {
      /* cannot parse response */

      MUFree(&S->QueueBuffer);

      return(FAILURE);
      }

    MUFree(&S->QueueBuffer);

    if (MXMLGetChild(E,"Response",NULL,&RE) == FAILURE)
      {
      /* cannot parse response */

      MXMLDestroyE(&E);

      return(FAILURE);
      }

    if (MXMLGetChild(RE,"Data",NULL,&DE) == FAILURE)
      {
      /* cannot parse response */

      MXMLDestroyE(&E);

      return(FAILURE);
      }
    }
  else
    {
    /* unexpected version */

    return(FAILURE);
    }

  CTok = -1;

  while(MXMLGetChild(DE,MSSSObjName[R->Version][mxoQueue],&CTok,&QE) == SUCCESS)
    {
    QID[0] = '\0';

    if ((__MSSSQueueGetState(QE,R,QID,&Status) == FAILURE) ||
        (MClassAdd(QID,&C) == FAILURE))
      {
      /* cannot add class */

      continue;
      }

    if ((Status == mclsDisabled) || (Status == mclsStopped))
      {
      C->IsDisabled = TRUE;
      }

    for (cindex = 0;cindex < QE->CCount;cindex++)
      {
      __MSSSQueueSetAttr(C,QE->C[cindex]->Name,QE->C[cindex]->Val);
      }  /* END for (cindex) */

    /* associate queues w/nodes */

    NL = (mnalloc_t *)C->NodeList;

    for (nindex = 0;nindex < MMAX_NODE;nindex++)
      {
      N = MNode[nindex];

      if ((N == NULL) || (N->Name[0] == '\0'))
        break;

      if (N->Name[0] == '\1')
        continue;

      if (NL != NULL)
        {
        int index;

        for (index = 0;NL[index].N != NULL;index++)
          {
          if (NL[index].N == N)
            break;
          }  /* END for (index) */

        if (NL[index].N == NULL)
          {
          /* node not in nodelist */

          continue;
          }
        }

      if (N->CRes.PSlot[C->Index].count == 0)
        {
        int CCount;

        CCount = N->CRes.Procs;

        if ((N->MaxProcPerClass != NULL) &&
            (N->MaxProcPerClass[C->Index] > 0))
          {
          CCount = MIN(CCount,N->MaxProcPerClass[C->Index]);
          }

        N->CRes.PSlot[C->Index].count += CCount;
        N->CRes.PSlot[0].count        += CCount;

        if (MClass[C->Index].IsDisabled != TRUE)
          {
          N->ARes.PSlot[C->Index].count += CCount;
          N->ARes.PSlot[0].count        += CCount;
          }
        }
      }  /* END for (nindex) */

    if (QCount != NULL)
      (*QCount)++;
    }    /* END while(MXMLGetChild(DE,MSSSObjName[R->Version][mxoQueue],&CTok,&QE) == SUCCESS) */

  MDB(1,fSSS) MLog("INFO:     all queues loaded\n");

  return(SUCCESS);
  }  /* END MSSSQueueQuery() */



 
int MSSSWorkloadQuery(

  mrm_t                *R,      /* I */
  int                  *JCount, /* O */
  char                 *EMsg,   /* O (optional) */
  enum MStatusCodeEnum *SC)     /* O (optional) */

  {
  mjob_t *J;
  
  enum MJobStateEnum Status;

  char   SJobID[MMAX_NAME];
  char   RMJID[MMAX_NAME];

  short  TaskList[MMAX_TASK_PER_JOB + 1];
  char   Message[MMAX_LINE];

  char  *ptr;

  sssrm_t *S;

  mxml_t *E = NULL;
  mxml_t *RE;
  mxml_t *DE;
  mxml_t *JE;

  mxml_t *StepE;
  mxml_t *NextSE;

  int      JCTok;
  int      SCTok;

  const char *FName = "MSSSWorkloadQuery";

  MDB(1,fSSS) MLog("%s(%s,JCount,EMsg,SC)\n",
    FName,
    (R != NULL) ? R->Name : "NULL");

  if (JCount != NULL)
    *JCount = 0;

  if ((R == NULL) || (R->S == NULL))
    {
    return(FAILURE);
    }

  S = (sssrm_t *)R->S;

  if (S->UpdateIteration != MSched.Iteration)
    {
    MSSSGetData(R,NULL);
    }
 
  /* step through list */
 
  if (S->WorkloadBuffer == NULL)
    {
    return(FAILURE);
    }

  if (R->Version == msssV2_0)
    {
    DE = (mxml_t *)S->WorkloadBuffer;

    if (DE == NULL)
      {
      return(FAILURE);
      }
    }
  else if (R->Version == msssV0_2)
    {
    if (((ptr = strstr(S->WorkloadBuffer,"<Body")) == NULL) ||
         (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
      {
      /* cannot parse response */
  
      MUFree(&S->WorkloadBuffer);
 
      return(FAILURE);
      }

    MUFree(&S->WorkloadBuffer);

    if (MXMLGetChild(E,(char *)MSON[msonResponse],NULL,&RE) == FAILURE)
      {
      /* cannot parse response */
 
      MXMLDestroyE(&E);
 
      return(FAILURE);
      }

    if (MXMLGetChild(RE,MSON[msonData],NULL,&DE) == FAILURE)
      {
      /* cannot parse response */

      MXMLDestroyE(&E);

      return(FAILURE);
      }
    }
  else
    {
    /* unexpected version */

    return(FAILURE);
    }

  JCTok = -1;

  while (MXMLGetChild(DE,MSSSObjName[R->Version][mxoJob],&JCTok,&JE) == SUCCESS)
    {
    /* load job default config */

    /* NYI */

    SCTok = -1;

    if (MXMLGetChild(JE,"Step",&SCTok,&NextSE) == FAILURE)
      {
      /* load single step job */

      NextSE = JE;
      }

    if (NextSE == NULL)
      {
      continue;
      }

    while (NextSE != NULL)
      {
      StepE = NextSE;

      MXMLGetChild(JE,"Step",&SCTok,&NextSE);

      /* NYI */

      RMJID[0] = '\0';

      if (__MSSSJobStepGetState(StepE,R,RMJID,&Status) == FAILURE)
        {
        /* job object is corrupt */

        MDB(1,fSSS) MLog("ALERT:    cannot determine jobid/state in %s (ignoring job)\n",
          FName);

        continue;
        }

      MJobGetName(NULL,RMJID,R,SJobID,MMAX_NAME,mjnShortName);

      if (JCount != NULL)
        (*JCount)++;

      switch (Status)
        {
        case mjsIdle:
        case mjsStarting:
        case mjsRunning:
        case mjsSuspended:
        case mjsHold:
 
          if (MJobFind(SJobID,&J,FALSE) == SUCCESS)
            {
            MRMJobPreUpdate(J);

            if (MSSSJobUpdate(StepE,&J,TaskList,R) == FAILURE)
              {
              MDB(1,fSSS) MLog("ALERT:    cannot update SSS job '%s'\n",
                SJobID);

              continue;
              }
 
            MRMJobPostUpdate(J,TaskList,Status,R);
            }
          else 
            {
            if (MJobCreate(SJobID,TRUE,&J) == FAILURE)
              {
              MDB(1,fSSS) MLog("ERROR:    job buffer is full  (ignoring job '%s')\n",
                SJobID);
 
              break;
              }

            /* if new job, load data */

            MRMJobPreLoad(J,SJobID,R->Index);

            MJobSetAttr(J,mjaRMJID,(void **)RMJID,mdfString,mSet);
 
            if (MSSSJobLoad(SJobID,StepE,&J,TaskList,R) == FAILURE)
              {
              MDB(1,fSSS) MLog("ALERT:    cannot load SSS job '%s'\n",
                SJobID);
 
              continue;
              }

            MRMJobPostLoad(J,TaskList,R);
            }

          MDB(2,fSSS)
            MJobShow(J,0,NULL);
 
          break;
 
        case mjsRemoved:
        case mjsCompleted:
        case mjsVacated:
 
          if (MJobFind(SJobID,&J,FALSE) == SUCCESS)
            {
            /* if job never ran, remove record.  job cancelled externally */
 
            if (MJOBISACTIVE(J) == FALSE)
              {
              MDB(1,fSSS) MLog("INFO:     job '%s' was cancelled externally\n",
                J->Name);
 
              /* remove job from joblist */
 
              MJOBREMOVE(&J);
 
              break;
              }

            MRMJobPreUpdate(J);
 
            if (MSSSJobUpdate(StepE,&J,TaskList,R) == FAILURE)
              {
              MDB(1,fSSS) MLog("ALERT:    cannot update SSS job '%s'\n",
                SJobID);

              continue;
              }
 
            MRMJobPostUpdate(J,TaskList,Status,R);

            MSysEMSubmit(
              &MSched.EM,
              (char *)MSSSCName[mpstSC],
              "job-terminated",
              J->Name);
 
            switch(Status)
              {
              case mjsRemoved:
              case mjsVacated:
 
                if (MSched.Time < (J->StartTime + J->WCLimit))
                  {
                  MJOBPROCESSREMOVED(&J);
                  }
                else
                  {
                  sprintf(Message,"JOBWCVIOLATION:  job '%s' exceeded WC limit %s\n",
                    J->Name,
                    MULToTString(J->WCLimit));
 
                  MSysRegEvent(Message,0,0,1);

                  MDB(3,fSSS) MLog("INFO:     job '%s' exceeded wallclock limit %s\n",
                    J->Name,
                    MULToTString(J->WCLimit));
 
                  MJOBPROCESSCOMPLETED(&J);
                  }
 
                break;
 
              case mjsCompleted:
 
                MJOBPROCESSCOMPLETED(&J);
 
                break;

              default:
 
                /* unexpected job state */
 
                MDB(1,fSSS) MLog("WARNING:  unexpected job state (%d) detected for job '%s'\n",
                  Status,
                  J->Name);
 
                break;
              }   /* END switch (Status)                        */
            }     /* END if (MJobFind(SJobID,&J,0) == SUCCESS) */
          else
            {
            /* ignore job */
 
            MDB(4,fSSS) MLog("INFO:     ignoring job '%s'  (state: %s)\n",
              SJobID,
              MJobState[Status]);
            }
 
          break;
 
        default:

          MDB(1,fSSS) MLog("WARNING:  job '%s' detected with unexpected state '%d'\n",
            SJobID,
            Status);

          /* NO-OP */
 
          break;
        }  /* END switch (Status) */
      }    /* END while(NextSE != NULL) */

    MDB(1,fSSS) MLog("INFO:     all steps loaded for job\n");
    }    /* END while (MXMLGetChild(DE,MSSSObjName[R->Version][mxoJob],&JCTok,&JE) == SUCCESS) */

  MDB(1,fSSS) MLog("INFO:     all jobs loaded\n");

  MXMLDestroyE(&E);

  /* TEMP:  remove jobs not detected via SSS */

  if (R->FailIteration != MSched.Iteration)
    {
    mjob_t *JNext;
    enum MNodeStateEnum OldState;

    for (J = MJob[0]->Next;(J != NULL) && (J != MJob[0]);J = JNext)
      {
      JNext = J->Next; /* store next job pointer in case current job is removed */

      if ((J->ATime > 0) &&
          (MSched.Time - J->ATime > MSched.JobPurgeTime))
        {
        if (MJOBISACTIVE(J) == TRUE)
          {
          MDB(1,fPBS) MLog("INFO:     active SSS job %s has been removed from the queue.  assuming successful completion\n",
            J->Name);

          MRMJobPreUpdate(J);

          /* assume job completed successfully for now */

          OldState          = J->State;

          J->State          = mjsCompleted;

          J->CompletionTime = J->ATime;

          MRMJobPostUpdate(J,NULL,OldState,J->SRM);

          MJOBPROCESSCOMPLETED(&J);
          }
        else
          {
          MDB(1,fPBS) MLog("INFO:     non-active SSS job %s has been removed from the queue.  assuming job was cancelled\n",
            J->Name);

          /* just remove job */

          MJOBREMOVE(&J);
          }
        }
      }    /* END for (jindex) */
    }      /* END if (R->FailIteration != MSched.Iteration) */

  return(SUCCESS);
  }  /* END MSSSWorkloadQuery() */




int __MSSSQueueGetState(

  mxml_t   *QE,      /* I */
  mrm_t    *R,       /* I */
  char     *QID,     /* O */
  int      *Status)  /* O */

  {
  mxml_t *CE;

  if (Status != NULL)
    *Status = mclsNONE;

  if ((QE == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  if ((QID != NULL) && (QID[0] == '\0'))
    {
    /* get queue name */

    if (MXMLGetChild(QE,MSSSQueueAttr[mclaName],NULL,&CE) == FAILURE)
      {
      return(FAILURE);
      }

    MUStrCpy(QID,CE->Val,MMAX_NAME);
    }  /* END if (QID != NULL) */

  if (Status != NULL)
    {
    if (MXMLGetChild(QE,MSSSQueueAttr[mclaState],NULL,&CE) == FAILURE)
      {
      /* class is available by default */

      *Status = mclsActive;
      }
    else
      {
      /* 
         mclsActive    accepting jobs, executing jobs 
         mclsDisabled  not accepting jobs, not executing jobs 
         mclsClosed    not accepting jobs, executing jobs 
         mclsStopped   accepting jobs, not executing jobs
      */

      if (MUBoolFromString(CE->Val,FALSE) == TRUE)
        {
        *Status = mclsActive;
        }
      else
        {
        *Status = mclsStopped;
        }
      }
    }    /* END if (Status != NULL) */

  return(SUCCESS);
  }  /* END __MSSSQueueGetState() */




int __MSSSJobStepGetState(

  mxml_t   *StepE,   /* I */
  mrm_t    *R,       /* I */
  char     *StepID,  /* O */
  enum MJobStateEnum *Status)  /* O */

  {
  mxml_t *XE;

  if (Status != NULL)
    *Status = mjsNONE;

  if ((StepE == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  if ((StepID != NULL) && (StepID[0] == '\0'))
    {
    switch(R->Version)
      {
      case msssV2_0:
 
        {
        mxml_t *StepIDE = NULL;

        if (MXMLGetChild(StepE,"StepId",NULL,&StepIDE) == SUCCESS)
          {
          MUStrCpy(StepID,StepIDE->Val,MMAX_NAME);
          }
        else if (MXMLGetChild(StepE,MSSSJobAttr[R->Version][mjaJobID],NULL,&StepIDE) == SUCCESS)
          {
          MUStrCpy(StepID,StepIDE->Val,MMAX_NAME);
          }
        else
          {
          /* cannot determine step ID */
   
          return(FAILURE);
          }
        }  /* END BLOCK */

        break;

      case msssV0_2:
      default:

        /* get job id */

        {
        mxml_t *C;

        if (MXMLGetChild(StepE,MSSSJobAttr[R->Version][mjaJobID],NULL,&C) == FAILURE)         
          {
          return(FAILURE);
          }

        MUStrCpy(StepID,C->Val,MMAX_NAME);
        }  /* END BLOCK */

        break;
      }  /* END switch(R->Version) */
    }    /* END if (StepID != NULL) */

  if (Status != NULL)
    {
    switch(R->Version)
      {
      case msssV2_0:

        {
        mxml_t *StateE = NULL;

        if (MXMLGetChild(StepE,"JobState",NULL,&StateE) == FAILURE)
          {
          return(FAILURE);
          }
    
        MUStrToLower(StateE->Val);
 
        *Status = MUGetIndex(StateE->Val,MSSSJobState,FALSE,0);
        }  /* END BLOCK */

        break;

      case msssV0_2:
      default:

        {
        XE = NULL;

        if (MXMLGetChild(StepE,MSSSJobAttr[R->Version][mjaState],NULL,&XE) == FAILURE)
          {
          mxml_t *C2;

          if (MXMLGetChild(StepE,MSSSJobAttr[R->Version][mjaAWDuration],NULL,&C2) == SUCCESS)
            { 
            if (!strcmp(C2->Val,"0"))
              *Status = mjsIdle;
            else
              *Status = mjsRunning;
            }
          else
            {
            *Status = mjsNotQueued;
            }
          }
        else
          {
          if ((*Status = MUGetIndex(XE->Val,MJobState,FALSE,0)) == 0)
            {
            *Status = MUGetIndex(XE->Val,MSSSJobState,FALSE,0);
            }
          }    /* END else (MXMLGetChild() == FAILURE) */
        }      /* END BLOCK */
      }        /* END switch(R->Version) */
    }          /* END if (Status != NULL) */

  return(SUCCESS);
  }  /* END __MSSSJobStepGetState() */





int MSSSClusterQuery(
 
  mrm_t *R,      /* I */
  int   *RCount, /* O */
  char  *EMsg,   /* O */
  enum MStatusCodeEnum *SC) /* I */
 
  {
  char  NodeName[MMAX_NAME];

  char *ptr;

  enum MNodeStateEnum Status;
  enum MNodeStateEnum OldState;

  mnode_t *N;

  sssrm_t *S;

  mbool_t  NewNode = FALSE;

  mxml_t  *RE = NULL;
  mxml_t  *DE = NULL;
  mxml_t  *NE = NULL;
  mxml_t  *E  = NULL;

  int      CTok;

  const char *FName = "MSSSClusterQuery";
    
  MDB(1,fSSS) MLog("%s(%s,RCount,SC)\n",
    FName,
    (R != NULL) ? R->Name : "NULL");
 
  if (RCount != NULL)
    *RCount = 0;
 
  if ((R == NULL) || (R->S == NULL))
    {
    MDB(1,fSSS) MLog("ALERT:    invalid parameters passed to %s\n",
      FName);

    return(FAILURE);
    }

  S = (sssrm_t *)R->S;

  /* step through list */

  if (S->UpdateIteration != MSched.Iteration)
    {
    MSSSGetData(R,NULL);
    }

  if (S->ResourceBuffer == NULL) 
    {
    MDB(1,fSSS) MLog("ALERT:    no resource data available in %s\n",
      FName);

    return(FAILURE);
    }

  if (R->Version == msssV2_0)
    {
    DE = (mxml_t *)S->ResourceBuffer;

    if (DE == NULL)
      {
      MDB(1,fSSS) MLog("ALERT:    invalid XML element passed to %s\n",
        FName);

      return(FAILURE);
      }
    }
  else if (R->Version == msssV0_2)
    {
    if (((ptr = strstr(S->ResourceBuffer,"<Body")) == NULL) ||
         (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
      {
      /* cannot parse response */

      MUFree(&S->ResourceBuffer);

      MDB(1,fSSS) MLog("ALERT:    cannot parse XML passed to %s\n",
        FName);

      return(FAILURE);
      }

    MUFree(&S->ResourceBuffer);

    if (MXMLGetChild(E,MSON[msonResponse],NULL,&RE) == FAILURE)
      {
      /* cannot parse response */

      MXMLDestroyE(&E);

      MDB(1,fSSS) MLog("ALERT:    cannot locate '%s' element in %s\n",
        MSON[msonResponse],
        FName);

      return(FAILURE);
      }

    if (MXMLGetChild(RE,MSON[msonData],NULL,&DE) == FAILURE)
      {
      /* cannot parse response */

      MXMLDestroyE(&E);

      MDB(1,fSSS) MLog("ALERT:    cannot locate '%s' element in %s\n",
        MSON[msonData],
        FName);

      return(FAILURE);
      }
    }
  else
    {
    /* unexpected version */

    MDB(1,fSSS) MLog("ALERT:    unexpected SSS version '%d' in %s\n",
      R->Version,
      FName);

    return(FAILURE);
    }

  CTok = -1;

  while(MXMLGetChild(DE,MSSSObjName[R->Version][mxoNode],&CTok,&NE) == SUCCESS)
    {
    NodeName[0] = '\0';
 
    if (__MSSSNodeGetState((void *)NE,R,NodeName,&Status) == FAILURE)
      {
      /* node object is corrupt */
 
      continue;
      }
 
    if (RCount != NULL)
      (*RCount)++;
 
    if (MNodeFind(NodeName,&N) == SUCCESS)
      {
      OldState = N->State;
 
      MRMNodePreUpdate(N,Status,R);
 
      MSSSNodeUpdate(N,NE,Status,R);
      }
    else if (MNodeAdd(NodeName,&N) == SUCCESS)
      {
      NewNode = TRUE;
 
      MRMNodePreLoad(N,Status,R);
 
      MSSSNodeLoad(N,NE,Status,R);
 
      MDB(2,fSSS)
        MNodeShow(N);
      }
    else
      {
      MDB(1,fSSS) MLog("ERROR:    node buffer is full  (ignoring node '%s')\n",
        NodeName);
      }
    }    /* END for (cindex) */

  /* all nodes loaded */

  MXMLDestroyE(&E);

  if (NewNode == TRUE)
    {
    MSSSQueueQuery(R,NULL,NULL,NULL);
    }
 
  return(SUCCESS);
  }  /* END MSSSClusterQuery() */





int MSSSQueueLoadInfo(
 
  mrm_t   *R,     /* I */
  mnode_t *SpecN) /* I (optional) */

  {
  const char *FName = "MSSSQueueLoadInfo";

  MDB(1,fSSS) MLog("%s(%s,%s)\n",
    FName,
    (R != NULL) ? R->Name : "NULL",
    (SpecN != NULL) ? SpecN->Name : "NULL");

  if (R == NULL)
    {
    return(FAILURE);
    }

  if ((MSched.Time > (R->U.SSS.ServerSDTimeStamp + 3000)) ||
      (R->U.SSS.ServerSD <= 0))
    {
    MSSSInitialize(R,NULL);

    if (R->U.SSS.ServerSD <= 0)
      {
      /* cannot recover SSS */

      MDB(1,fSSS) MLog("ALERT:    cannot re-initialize SSS interface\n");

      return(FAILURE);
      }
    }

  return(SUCCESS);
  }  /* END MSSSQueueLoadInfo() */





int __MSSSNodeGetState(

  void  *NP,     /* I */
  mrm_t *R,      /* I */
  char  *NodeID, /* O */
  enum MNodeStateEnum *Status) /* O */

  {
  mxml_t *E;

  const char *FName = "__MSSSNodeGetState";

  MDB(1,fSSS) MLog("%s(%s,%s,%s,%s)\n",
    FName,
    (NP != NULL) ? "NP" : "NULL",
    (R != NULL) ? R->Name : "NULL",
    (NodeID != NULL) ? "NodeID" : "NULL",
    (Status != NULL) ? "Status" : "NULL");

  if ((NP == NULL) || (R == NULL))
    {
    return(FAILURE);
    }
 
  E = (mxml_t *)NP;
 
  if ((NodeID != NULL) && (NodeID[0] == '\0'))
    {
    /* get node name */
 
    mxml_t *C;
 
    if (MXMLGetChild(E,MSSSNodeAttr[R->Version][mnaNodeID],NULL,&C) == FAILURE)
      {
      return(FAILURE);
      }

    strcpy(NodeID,C->Val); 
    }

  if (Status != NULL)
    {
    mxml_t *C = NULL;

    *Status = mnsNone;

    if (MXMLGetChild(E,MSSSNodeAttr[R->Version][mnaNodeState],NULL,&C) == FAILURE)
      {
      /* node state not specified */

      *Status = mnsUnknown;      
      }
    else if ((R->Version == msssV2_0) ||
        (R->Version == msssV0_2))
      {
      const char *SSSNState[] = { NONE, "up", "down", NULL };
      const enum MNodeStateEnum SSSNSEq[] = { mnsNone, mnsUnknown, mnsDown, mnsNONE };

      int index;

      /* translate SSS node state to internal node state */

      index = MUGetIndex(C->Val,SSSNState,FALSE,mnsNONE);

      *Status = SSSNSEq[index];
      }
    else
      {
      const char *SSSNState[] = { NONE, "active", "busy", "down", "idle", NULL };
      const enum MNodeStateEnum SSSNSEq[] = { mnsNone, mnsActive, mnsBusy, mnsDown, mnsIdle, mnsNONE };

      int index;

      /* translate SSS node state to internal node state */

      index = MUGetIndex(C->Val,SSSNState,FALSE,mnsNONE);

      *Status = SSSNSEq[index];
      }  /* END else () */

    MDB(1,fSSS) MLog("INFO:     SSS node %s set to state %s (%s)\n",
      (NodeID != NULL) ? NodeID : NONE,
      MNodeState[*Status],
      (C != NULL) ? C->Val : "NULL");
    }    /* END if (Status != NULL) */
    
  return(SUCCESS);
  }  /* END __MSSSNodeGetState() */





int MSSSJobStart(

  mjob_t *J,   /* I */
  mrm_t  *R,   /* I */
  char   *Msg, /* O (optional) */
  int    *SC)  /* O (optional) */

  {
  char         Message[MMAX_LINE];

  char        *Response = NULL;

  char         CmdString[MMAX_BUFFER];

  int          nindex;
  int          tindex;

  char        *MasterHost;

  char        *ptr;

  mnode_t *N;

  mxml_t *E  = NULL;
  mxml_t *RE = NULL;
  mxml_t *DE;
  mxml_t *JE;
  mxml_t *NE;
  mxml_t *C  = NULL;

  char       tmpMsg[MMAX_LINE];
  enum MSFC  tmpSC;

  const char *FName = "MSSSJobStart";

  MDB(1,fSSS) MLog("%s(%s,%s,Msg,SC)\n",
    FName,
    (J != NULL) ? J->Name : "NULL",
    (R != NULL) ? R->Name : "NULL");

  if (SC != NULL)
    *SC = mscNoError;

  if ((J == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  MasterHost  = NULL;

  if (J->NodeList[0].N == NULL)
    {
    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be started: (empty hostlist)\n",
      J->Name);
 
    return(FAILURE);
    }

  switch(R->Version)
    {
    case msssV2_0:


      MXMLCreateE(&E,"Request");
      MXMLSetAttr(E,"action",(void *)MSSSJAName[R->Version][mrmJobStart],mdfString);

      MSSSSetObject(E,MSSSObjName[R->Version][mxoJob],NULL);


      DE = NULL;
      MXMLCreateE(&DE,"Data");
      MXMLAddE(E,DE);

      JE = NULL;
      MXMLCreateE(&JE,MSSSObjName[R->Version][mxoJob]);
      MXMLAddE(DE,JE);

      C = NULL;
      MXMLCreateE(&C,MSSSJobAttr[R->Version][mjaJobID]); 
      /* MXMLCreateE(&C,"StepId"); */
      MXMLSetVal(C,(void *)J->Name,mdfString);
      MXMLAddE(JE,C);

      C = NULL;
      MXMLCreateE(&C,"NodeList");
      MXMLAddE(JE,C);
   
      for (nindex = 0;J->NodeList[nindex].N != NULL;nindex++)
        {
        if (J->NodeList[nindex].N->RM != R)
          continue;

        for (tindex = 0;tindex < MAX(1,J->NodeList[nindex].TC);tindex++)
          {
          N = J->NodeList[nindex].N;

          NE = NULL;
          MXMLCreateE(&NE,MSSSObjName[R->Version][mxoNode]);
          MXMLSetVal(NE,(void *)N->Name,mdfString);
          MXMLAddE(C,NE);
          }  /* END for (tindex) */
        }    /* END for (tindex) */

      break;

    case msssV0_2:

      MXMLCreateE(&E,"Message");
      MXMLCreateE(&RE,"Request");
 
      MXMLAddE(E,RE);

      MXMLSetAttr(RE,"action",(void *)"start",mdfString);
      MXMLSetAttr(RE,"object",(void *)"job",mdfString);

      C = NULL;
      MXMLCreateE(&C,(char *)MSON[msonWhere]);
      MXMLSetAttr(C,"name",(void *)MSSSJobAttr[R->Version][mjaJobID],mdfString);
      MXMLSetAttr(C,"value",(void *)J->Name,mdfString);
      MXMLAddE(RE,C);

      for (nindex = 0;J->NodeList[nindex].N != NULL;nindex++)
        {
        if (J->NodeList[nindex].N->RM != R)
          continue;

        for (tindex = 0;tindex < MAX(1,J->NodeList[nindex].TC);tindex++)
          {
          N = J->NodeList[nindex].N;
  
          C = NULL;
          MXMLCreateE(&C,(char *)MSON[msonSet]);
          MXMLSetAttr(C,"name",(void *)"node",mdfString);
          MXMLSetAttr(C,"value",(void *)N->Name,mdfString);
          MXMLAddE(RE,C);
          }  /* END for (tindex) */
        }    /* END for (tindex) */

      break;

    default:

      /* not supported */

      return(FAILURE);
    
      /*NOTREACHED*/

      break;
    }      /* END switch(R->Version) */

  MXMLToString(E,CmdString,sizeof(CmdString),NULL,TRUE);
 
  MXMLDestroyE(&E);

  if (MSSSDoCommand(
        &R->P[0],
        CmdString,
        &Response,
        NULL) == FAILURE)
    {
    /* job could not be started */

    if (Response != NULL)
      {
      char tmpLine[MMAX_LINE];

      snprintf(tmpLine,sizeof(tmpLine),"cannot start job - %s",
        Response);

      free(Response);

      MUStrDup(&J->Message,tmpLine);
      }

    R->FailIteration = MSched.Iteration;

    if (SC != NULL)
      *SC = mscRemoteFailure;

    MSUFree(R->P[0].S);

    return(FAILURE);
    }

  if (Response != NULL)
    {
    if (((ptr = strstr(Response,"<Body")) == NULL) ||
         (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
      {
      /* cannot parse response */
 
      MUFree(&Response);

      R->FailIteration = MSched.Iteration;        

      MSUFree(R->P[0].S);
 
      return(FAILURE);
      }

    MUFree(&Response);     

    if (MXMLGetChild(E,"Response",NULL,&RE) == FAILURE)
      {
      /* cannot parse response */
 
      MXMLDestroyE(&E);

      R->FailIteration = MSched.Iteration;  
      }
    }

  /* check status code */

  if (MSSSCheckStatus(RE,&tmpSC,tmpMsg) == FAILURE)
    {
    MXMLDestroyE(&E);

    R->FailIteration = MSched.Iteration;

    snprintf(Message,MMAX_LINE,"cannot start job - RM failure, rc: %d, msg: '%s'",
      tmpSC,
      tmpMsg);

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be started: RM failure (%s)\n",
      J->Name,
      Message);

    if (Msg != NULL)
      strcpy(Msg,Message);

    MUStrDup(&J->Message,Message);

    MSUFree(R->P[0].S);

    return(FAILURE);
    }

  /* job successfully started */

  MXMLDestroyE(&E);      

  /* NOTE: SSS does not provide accurate job start info in many cases */

  J->StartTime    = MSched.Time;
  J->DispatchTime = MSched.Time;

  MSysEMSubmit(
    &MSched.EM,
    (char *)MSSSCName[mpstSC],
    (char *)"jobstart",
    (char *)J->Name);

  MDB(1,fSSS) MLog("INFO:     job '%s' successfully started\n",
    J->Name);

  return(SUCCESS);
  }  /* END MSSSJobStart() */




int MSSSJobCancel(

  mjob_t *J,        /* I */
  mrm_t  *R,        /* I */
  char   *Message,  /* I */
  char   *EMsg,     /* O (optional) */
  int    *SC)       /* O (optional) */

  {
  mxml_t *E  = NULL;
  mxml_t *RE = NULL;
  mxml_t *CE = NULL;
  mxml_t *DE = NULL;
  mxml_t *JE = NULL;

  char       tmpLine[MMAX_LINE];
  char       tmpMsg[MMAX_LINE];
  enum MSFC  tmpSC;

  char *ptr;

  char *Response = NULL; 

  char  CmdString[MMAX_BUFFER];    

  const char *FName = "MSSSJobCancel";

  MDB(1,fSSS) MLog("%s(%s,%s,%s,EMsg,SC)\n",
    FName,
    (J != NULL) ? J->Name : "NULL",
    (R != NULL) ? R->Name : "NULL",
    (Message != NULL) ? Message : "NULL");
 
  if ((J == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  switch(R->Version)
    {
    case msssV2_0:

      MXMLCreateE(&E,MSON[msonRequest]);
      MXMLSetAttr(E,MSAN[msanAction],(void *)MSSSJAName[R->Version][mrmJobCancel],mdfString);

      MSSSSetObject(E,MSSSObjName[R->Version][mxoJob],NULL);

      DE = NULL;
      MXMLCreateE(&DE,MSON[msonData]);
      MXMLAddE(E,DE);

      JE = NULL;
      MXMLCreateE(&JE,MSSSObjName[R->Version][mxoJob]);
      MXMLAddE(DE,JE);

      CE = NULL;
      MXMLCreateE(&CE,MSSSJobAttr[R->Version][mjaJobID]);
      /* MXMLCreateE(&CE,"StepId"); */
      MXMLSetVal(CE,(void *)J->Name,mdfString);
      MXMLAddE(JE,CE);

      break;
    }  /* END switch(R->Version) */
 
  MXMLToString(E,CmdString,sizeof(CmdString),NULL,TRUE);
 
  MXMLDestroyE(&E);
 
  if (MSSSDoCommand(
        &R->P[0],
        CmdString,
        &Response,
        NULL) == FAILURE)
    {
    /* job could not be cancelled */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be cancelled (command failed)\n",
      J->Name);
 
    R->FailIteration = MSched.Iteration;
 
    return(FAILURE);
    }

  if (((ptr = strstr(Response,"<Body")) == NULL) ||
       (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
    {
    /* cannot parse response */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be cancelled (invalid response)\n",
      J->Name);

    MUFree(&Response);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  MUFree(&Response);
 
  if (MXMLGetChild(E,"Response",NULL,&RE) == FAILURE)
    {
    /* cannot parse response */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be cancelled (cannot parse response)\n",
      J->Name);

    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;
    }

  /* check status code */

  if (MSSSCheckStatus(RE,&tmpSC,tmpMsg) == FAILURE)
    {
    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    snprintf(tmpLine,MMAX_LINE,"cannot cancel job - RM failure, rc: %d, msg: '%s'",
      tmpSC,
      tmpMsg);

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be cancelled: RM failure (%s)\n",
      J->Name,
      tmpMsg);

    if (EMsg != NULL)
      strcpy(EMsg,tmpLine);

    MUStrDup(&J->Message,tmpLine);

    MSUFree(R->P[0].S);

    return(FAILURE);
    }
 
  /* job successfully cancelled */

  MSysEMSubmit(
    &MSched.EM,
    (char *)MSSSCName[mpstSC],
    (char *)"jobcancel",
    (char *)J->Name);

  MDB(1,fSSS) MLog("INFO:     job '%s' successfully cancelled\n",
    J->Name);

  return(SUCCESS);
  }  /* END MSSSJobCancel() */




int MSSSNodeLoad(

  mnode_t   *N,      /* I (update) */
  void      *NP,     /* I */
  int        Status, /* I */
  mrm_t     *R)      /* I */

  {
  int           nindex;
  int           cindex;

  mulong        tmpTime;

  mxml_t      *E;
  mxml_t      *C;

  const char *FName = "MSSSNodeLoad";

  MDB(2,fSSS) MLog("%s(%s,%s,%s,%s)\n",
    FName,
    (N != NULL) ? N->Name : "NULL",
    (NP != NULL) ? "NP" : "NULL",
    MNodeState[Status],
    (R != NULL) ? R->Name : "NULL");

  if ((N == NULL) || (NP == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  MTRAPNODE(N,FName);

  MUGetTime(&tmpTime,mtmNONE,NULL);

  N->CTime = tmpTime;
  N->MTime = tmpTime;
  N->ATime = tmpTime;

  N->RM    = R;

  N->TaskCount = 0;

  if (N->ARes.Disk > 0)
    {
    if (N->CRes.Disk <= 0)
      N->CRes.Disk = N->ARes.Disk;
    }
  else
    {
    N->ARes.Disk = 1;
    N->CRes.Disk = 1;
    }

  if (N->CRes.Mem > 0)
    {
    if (N->ARes.Mem <= 0)
      N->ARes.Mem = N->CRes.Mem;
    }
  else
    {
    N->ARes.Mem = 1;
    N->CRes.Mem = 1;
    }

  if (N->CRes.Swap > 0)
    {
    if (N->ARes.Swap <= 0)
      N->ARes.Swap = N->CRes.Swap;
    }
  else
    {
    /* virtual memory always at least as large as real memory */

    N->ARes.Swap = MAX(MDEF_MINSWAP,N->ARes.Mem);
    N->CRes.Swap = MAX(MDEF_MINSWAP,N->CRes.Mem);
    }

  N->ActiveOS = MUMAGetIndex(meOpsys,"DEFAULT",mAdd);

  N->Network = MUMAGetBM(meNetwork,"DEFAULT",mAdd);

  N->TaskCount = 0;

  MLocalNodeInit(N);

  E = (mxml_t *)NP;
 
  for (cindex = 0;cindex < E->CCount;cindex++)
    {
    C = E->C[cindex];

    __MSSSNodeSetAttr(N,R,C->Name,C->Val,(void *)C);
    }  /* END for (cindex) */

  for (cindex = 0;cindex < MMAX_CLASS;cindex++)
    {
    mclass_t *C;
    mnalloc_t *NL;

    C = &MClass[cindex];

    if ((C == NULL) || (C->Name[0] == '\0'))
      break;

    if (!strcmp(C->Name,"DEFAULT") || !strcmp(C->Name,"ALL"))
      continue;

    if (C->NodeList != NULL)
      {
      NL = (mnalloc_t *)C->NodeList;

      for (nindex = 0;nindex < MMAX_NODE;nindex++)
        {
        if (NL[nindex].N == NULL)
          break;
  
        if (NL[nindex].N == N)
          break; 
        }  /* END for (nindex) */      

      if ((nindex >= MMAX_NODE) || (NL[nindex].N == NULL))
        continue;
      }  /* END if (C->NodeList != NULL) */

    MNodeSetClass(N,C,NULL,mAdd);
    }  /* END for (cindex) */

  MDB(6,fSSS) MLog("INFO:     MNode[%03d] '%18s' %9s VM: %8d Mem: %5d Dk: %5d Cl: %6s %s\n",
    N->Index,
    N->Name,
    MAList[meNodeState][N->State],
    N->CRes.Swap,
    N->CRes.Mem,
    N->CRes.Disk,
    MUNumListToString(N->ARes.PSlot,N->CRes.PSlot,NULL,NULL,0),
    MUMAToString(meNFeature,',',N->FBM,sizeof(N->FBM)));

  MDB(6,fSSS) MLog("INFO:     MNode[%03d] '%18s' C/A/D procs:  %d/%d/%d\n",
    N->Index,
    N->Name,
    N->CRes.Procs,
    N->ARes.Procs,
    N->DRes.Procs);

  return(SUCCESS);
  }  /* END MSSSNodeLoad() */






int MSSSNodeUpdate(

  mnode_t   *N,      /* I (modified) */
  void      *NP,     /* I */
  int        Status, /* I */
  mrm_t     *R)      /* I */

  {
  int           cindex;

  mulong        tmpTime;

  mxml_t      *E;
  mxml_t      *C;

  const char *FName = "MSSSNodeUpdate";

  MDB(2,fSSS) MLog("%s(%s,%s,%s,%s)\n",
    FName,
    (N != NULL) ? N->Name : "NULL",
    (NP != NULL) ? "NP" : "NULL",
    MNodeState[Status],
    (R != NULL) ? R->Name : "NULL");
 
  if ((N == NULL) || (NP == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  MUGetTime(&tmpTime,mtmNONE,NULL);

  N->MTime   = tmpTime;
  N->ATime   = tmpTime;

  N->RM      = R;

  N->TaskCount = 0;

  if (N->ARes.Disk > 0)
    {
    if (N->CRes.Disk <= 0)
      N->CRes.Disk = N->ARes.Disk;
    }
  else
    {
    N->ARes.Disk = 1;
    N->CRes.Disk = 1;
    }

  if (N->CRes.Mem > 0)
    {
    if (N->ARes.Mem <= 0)
      N->ARes.Mem = N->CRes.Mem;
    }
  else
    {
    N->ARes.Mem = 1;
    N->CRes.Mem = 1;
    }

  if (N->CRes.Swap > 0)
    {
    if (N->ARes.Swap <= 0)
      N->ARes.Swap = N->CRes.Swap;
    }
  else
    {
    /* virtual memory always at least as large as real memory */

    N->ARes.Swap = MAX(MDEF_MINSWAP,N->ARes.Mem);
    N->CRes.Swap = MAX(MDEF_MINSWAP,N->CRes.Mem);
    }

  /* SSS does not provide pool, opsys, machine speed, or network info */

  N->ActiveOS = MUMAGetIndex(meOpsys,"DEFAULT",mAdd);

  N->Network = MUMAGetBM(meNetwork,"DEFAULT",mAdd);

  N->Load    = 0.0;

  /* get joblist, maxtask, and feature info from SSS server */

  N->TaskCount = 0;

  E = (mxml_t *)NP;

  for (cindex = 0;cindex < E->CCount;cindex++)
    {
    C = E->C[cindex];

    __MSSSNodeSetAttr(N,R,C->Name,C->Val,(void *)C);
    }  /* END for (cindex) */
 
  MDB(6,fSSS) MLog("INFO:     MNode[%03d] '%18s' %9s VM: %8d Mem: %5d Dk: %5d Cl: %6s %s\n",
    N->Index,
    N->Name,
    MAList[meNodeState][N->State],
    N->CRes.Swap,
    N->CRes.Mem,
    N->CRes.Disk,
    MUNumListToString(N->ARes.PSlot,N->CRes.PSlot,NULL,NULL,0),
    MUMAToString(meNFeature,',',N->FBM,sizeof(N->FBM)));

  MDB(6,fSSS) MLog("INFO:     MNode[%03d] '%18s' C/A/D procs:  %d/%d/%d\n",
    N->Index,
    N->Name,
    N->CRes.Procs,
    N->ARes.Procs,
    N->DRes.Procs);

  return(SUCCESS);
  }  /* END MSSSNodeUpdate() */




int MSSSJobLoad(

  char    *JobName,  /* I/O */
  void    *SJP,      /* I (XML) */
  mjob_t **JP,       /* I (modified/freed on failure) */
  short   *TaskList, /* O */
  mrm_t   *R)        /* I */

  {
  mreq_t       *RQ;

  mqos_t       *QDef;

  int           cindex;

  long          WallTime;
  long          WCLimit;

  int           NodeCount;
  int           TaskCount;

  mxml_t       *E;
  mxml_t       *C;

  mjob_t       *J;

  char          tmpLine[MMAX_LINE];

  const char *FName = "MSSSJobLoad";

  MDB(2,fSSS) MLog("%s(%s,SJP,%s,TaskList,%s)\n",
    FName,
    (JobName != NULL) ? JobName : "NULL",
    ((JP != NULL) && (*JP != NULL)) ? (*JP)->Name : "NULL",
    (R != NULL) ? R->Name : "NULL");
 
  if ((SJP == NULL) || (JP == NULL) || (*JP == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  if ((JobName == NULL) && !MISSET(R->Flags,mrmfLocalQueue))
    {
    return(FAILURE);
    }

  /* NOTE:  only assign job id if job successfully processed */

  J = *JP;
 
  MTRAPJOB(J,FName);

  WallTime = 0;
  WCLimit  = 0;

  TaskCount = 1;
  NodeCount = 1;

  /* add resource requirements information */

  if (MReqCreate(J,NULL,&RQ,FALSE) == FAILURE)
    {
    MDB(1,fSSS) MLog("ALERT:    cannot add requirements to job '%s'\n",
      J->Name);

    MJOBREMOVE(JP);

    return(FAILURE);
    }

  MRMReqPreLoad(RQ);

  TaskList[0]    = -1;

  J->Request.TC  = 0;

  J->Request.NC  = 0;
  RQ->NodeCount  = 0;

  RQ->DRes.Procs = 1;
  RQ->DRes.Mem   = 0;

  RQ = J->Req[0];  /* FIXME:  only allow one req per job in SSS */

  E = (mxml_t *)SJP;
 
  for (cindex = 0;cindex < E->CCount;cindex++)
    {
    C = E->C[cindex];

    /* check units */

    if (MXMLGetAttr(C,"units",NULL,tmpLine,0) == SUCCESS)
      {
      /* NYI */
      }

    __MSSSJobSetAttr(J,R->Version,C->Name,C->Val,MBNOTSET,C);
    }  /* END for (cindex) */

  if (J->Request.TC == 0)
    {
    MDB(2,fSSS) MLog("ALERT:    no job task info located for job '%s' (assigning taskcount to 1)\n",
      J->Name);

    J->Request.TC = MAX(1,J->Request.NC);
    RQ->TaskCount = MAX(1,J->Request.NC);
    }

  if ((RQ->URes.Procs > 0) && (RQ->TaskCount > 0))
    {
    RQ->URes.Procs /= RQ->TaskCount;
    }

  /* authenticate job */

  if (J->Cred.U == NULL)
    {
    MDB(1,fSSS) MLog("ERROR:    cannot authenticate job '%s' (User: %s  Group: %s)\n",
      J->Name,
      (J->Cred.U != NULL) ? J->Cred.U->Name : "NULL",
      (J->Cred.G != NULL) ? J->Cred.G->Name : "NULL");

    MJOBREMOVE(JP);

    return(FAILURE);
    }

  if (MJobSetCreds(
        J,
        J->Cred.U->Name,
        DEFAULT,
        (J->Cred.A != NULL) ? J->Cred.A->Name : NULL) == FAILURE)
    {
    MDB(1,fSSS) MLog("ERROR:    cannot authenticate job '%s' (U: %s  G: %s  A: '%s')\n",
      J->Name,
      J->Cred.U->Name,
      J->Cred.G->Name,
      (J->Cred.A != NULL) ? J->Cred.A->Name : NONE);

    MJOBREMOVE(JP);

    return(FAILURE);
    }

  if ((MQOSGetAccess(J,J->QReq,NULL,&QDef) == FAILURE) ||
      (J->QReq == NULL))
    {
    MJobSetQOS(J,QDef,0);
    }
  else
    {
    MJobSetQOS(J,J->QReq,0);
    }

  /* set defaults for info not specified by SSS */

  J->SubmitTime = MIN(J->SubmitTime,MSched.Time);

  RQ = J->Req[0];

  RQ->RMIndex    = R->Index;

  RQ->TaskRequestList[0] = J->Request.TC;
  RQ->TaskRequestList[1] = J->Request.TC;

  RQ->TaskRequestList[2] = 0;

  /* adjust resource requirement info */
  /* NOTE:  SSS specifies resource requirement info on a 'per node' basis */
 
  RQ->DRes.Mem  /= RQ->TaskCount;
  RQ->DRes.Swap /= RQ->TaskCount;
  RQ->DRes.Disk /= RQ->TaskCount;     

  J->ExecSize  = 0;
  J->ImageSize = 0;

  RQ->NAccessPolicy = MSched.DefaultNAccessPolicy;

  MUStrDup(&J->E.IWD,NONE);
  MUStrDup(&J->E.Cmd,NONE);

  if ((J->State == mjsStarting) ||
      (J->State == mjsRunning) ||
      (J->State == mjsCompleted) ||
      (J->State == mjsRemoved))
    {
    /* obtain 'per task' statistics */

    if (RQ->TaskCount > 0)
      {  
      if (RQ->URes.Procs > 0)
        RQ->URes.Procs /= RQ->TaskCount;
      }
    }

  if (JobName == NULL)
    {
    /* set job name */

    /* need rm counter */

    /* NYI */
    }

  MLocalJobInit(J);

  return(SUCCESS);
  }  /* END MSSSJobLoad() */




int MSSSJobUpdate(

  void      *SJP,      /* I */
  mjob_t   **JP,       /* I (modified/freed on failure) */
  short     *TaskList, /* I */
  mrm_t     *R)        /* I */

  {
  enum MNodeStateEnum OldState;

  int           TaskCount;
  int           NodeCount;

  mreq_t        *RQ;

  int           cindex;

  mxml_t       *E;
  mxml_t       *C;

  char          tmpLine[MMAX_LINE];

  mjob_t       *J;

  const char *FName = "MSSSJobUpdate";

  MDB(2,fSSS) MLog("%s(SJP,%s,TaskList,%s)\n",
    FName,
    ((JP != NULL) && (*JP != NULL)) ? (*JP)->Name : "NULL",
    (R != NULL) ? R->Name : "NULL");

  if ((JP == NULL) || (*JP == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  J = *JP;

  OldState = J->State;

  TaskList[0] = -1;

  /* get job state: NYI */

  RQ = J->Req[0];

  TaskCount = 1;
  NodeCount = 1;

  if (J->State != OldState)
    {
    MDB(1,fSSS) MLog("INFO:     job '%s' changed states from %s to %s\n",
      J->Name,
      MJobState[OldState],
      MJobState[J->State]);

    J->MTime = MSched.Time;
    }

  E = (mxml_t *)SJP;

  for (cindex = 0;cindex < E->CCount;cindex++)
    {
    C = E->C[cindex];

    /* check units */

    if (MXMLGetAttr(C,"units",NULL,tmpLine,0) == SUCCESS)
      {
      /* NYI */
      }

    __MSSSJobSetAttr(J,R->Version,C->Name,C->Val,MBNOTSET,C);
    }  /* END for (cindex) */

  if (J->State != OldState)
    {
    MDB(1,fSSS) MLog("INFO:     job '%s' changed states from %s to %s\n",
      J->Name,
      MJobState[OldState],
      MJobState[J->State]);

    J->MTime = MSched.Time;
    }

  if ((J->State == mjsCompleted) ||
      (J->State == mjsRemoved) ||
      (J->State == mjsVacated))
    {
    /* NOTE:  no completion time attribute currently supported */

    J->CompletionTime = J->MTime;

    if (J->AWallTime > 0)
      J->CompletionTime = MIN(J->CompletionTime,J->StartTime + J->AWallTime);
    }

  return(SUCCESS);
  }  /* END MSSSJobUpdate() */




int __MSSSNodeSetAttr(

  mnode_t *N,     /* I (modified) */
  mrm_t   *R,     /* I */
  char    *AName, /* I */
  char    *AVal,  /* I */
  void    *AData) /* I (optional) */
 
  {
  int aindex;

  long tmpL;

  char UData[MMAX_NAME];
  char tmpData[MMAX_NAME];

  if ((N == NULL) || (AName == NULL))
    {
    return(FAILURE);
    }

  if ((aindex = MUGetIndex(AName,(const char **)MSSSNodeAttr[R->Version],FALSE,0)) != 0)
    {
    if (AVal == NULL)
      {
      return(SUCCESS);
      }

    switch(aindex)
      {
      case mnaRCProc:
      case mnaRAProc:
      case mnaRCSwap:
      case mnaRASwap:
      case mnaRCDisk:
      case mnaRADisk:
      case mnaRCMem:
      case mnaRAMem:

        tmpL = -1;

        if (AData != NULL)
          {
          /* check units */

          if (MXMLGetAttr((mxml_t *)AData,"units",NULL,UData,sizeof(UData)) == SUCCESS)
            {
            sprintf(tmpData,"%s%s",
              AVal,
              UData);

           /* adjust value by unit information */

           tmpL = MURSpecToL(tmpData,mvmMega,mvmMega);
           }
         }

        if (tmpL == -1)
          tmpL = MURSpecToL(AVal,mvmMega,mvmMega);

      case mnaNodeID:
      case mnaLoad:

        if (strtol(AVal,NULL,0) > 0)
          MNodeSetAttr(N,aindex,(void **)AVal,mdfString,mSet);

        break;

      case mnaNodeState:

        if (!strcmp(R->APIVersion,MSSSVName[msssV2_0]) ||
            !strcmp(R->APIVersion,MSSSVName[msssV0_2]))
          {
          const char *SSSNState[] = { NONE, "up", "down", NULL };
          const int   SSSNSEq[]   = { mnsNone, mnsUnknown, mnsDown, -1 };

          int index;
          int SIndex;

          /* translate SSS node state to internal node state */

          index = MUGetIndex(AVal,SSSNState,FALSE,0);

          SIndex = SSSNSEq[index];

          MNodeSetAttr(N,mnaNodeState,(void **)&SIndex,mdfInt,mSet);
          }
        else
          {
          const char *SSSNState[] = { NONE, "active", "busy", "down", "idle", NULL };
          const int   SSSNSEq[]   = { mnsNone, mnsActive, mnsBusy, mnsDown, mnsIdle, -1 };

          int index;
          int SIndex;

          /* translate SSS node state to M node state */

          index = MUGetIndex(AVal,SSSNState,FALSE,0);

          SIndex = SSSNSEq[index];

          MNodeSetAttr(N,mnaNodeState,(void **)&SIndex,mdfInt,mSet);
          }  /* END else () */

        if (N->State == mnsUnknown)
          {
          N->ARes.Procs = -1;
          }

        break;

      default:

        /* NO-OP */

        break;
      }  /* END switch(aindex) */
    }    /* END if (MUGetIndex(AName,(const char **)MSSSNodeAttr[R->Version],FALSE,0) != 0) */
  else
    {
    int rindex;

    mxml_t *AE;
    mxml_t *CE;

    mbool_t IsConfigured;

    int CTok;

    /* unrecognized attribute */

    AE = (mxml_t *)AData;

    if (strcmp(AName,"Available"))
      {
      IsConfigured = TRUE;
      }
    else
      {
      IsConfigured = FALSE;
      }

    CTok = -1;

    while (MXMLGetChild(AE,NULL,&CTok,&CE) == SUCCESS)
      {
      AVal = CE->Val;

      if ((rindex = MUGetIndex(CE->Name,(const char **)MSSSNodeRAttr[R->Version],FALSE,0)) == 0)
        {
        return(FAILURE);
        }

      tmpL = -1;

      /* check units */

      if (MXMLGetAttr((mxml_t *)AData,"units",NULL,UData,sizeof(UData)) == SUCCESS)
        {
        sprintf(tmpData,"%s%s",
          AVal,
          UData);

        /* adjust value by unit information */

         tmpL = MURSpecToL(tmpData,mvmMega,mvmMega);
         }

      if (tmpL == -1)
        tmpL = MURSpecToL(AVal,mvmMega,mvmMega);

      switch(rindex)
        {
        case mrDisk:

          aindex = (IsConfigured == TRUE) ? mnaRCDisk : mnaRADisk;
  
          break;

        case mrMem:

          aindex = (IsConfigured == TRUE) ? mnaRCMem : mnaRAMem;

          break;

        case mrProc:

          aindex = (IsConfigured == TRUE) ? mnaRCProc : mnaRAProc;

          break;

        case mrSwap:

          aindex = (IsConfigured == TRUE) ? mnaRCSwap : mnaRASwap;

          break;
  
        default:

          /* NOT HANDLED */

          return(FAILURE);
       
          /*NOTREACHED*/

          break;
        }  /* END switch(rindex) */

      MNodeSetAttr(N,aindex,(void **)AVal,mdfString,mSet);
      }  /* END while (MXMLGetChild(CE,NULL,&CTok,&CE) == SUCCCESS) */
    }    /* END else ... */

  return(SUCCESS);
  }  /* END __MSSSNodeSetAttr() */




int __MSSSQueueSetAttr(

  mclass_t *C,     /* I (modified) */
  char     *AName, /* I */
  char     *AVal)  /* I */

  {
  int aindex;

  if ((C == NULL) || (AName == NULL))
    {
    return(FAILURE);
    }

  if ((aindex = MUGetIndex(AName,(const char **)MSSSQueueAttr,FALSE,0)) != 0)
    {
    switch(aindex)
      {
      case mclaHostList:

        {
        char *ptr;
        char *TokPtr;

        mnode_t *N;
        mnalloc_t *NL;

        int   nindex;

        if (C->NodeList == NULL)
          {
          if ((C->NodeList = (void *)calloc(
               1,
               sizeof(mnalloc_t) * MMAX_NODE)) == NULL)
            {
            break;
            }
          }

        NL = (mnalloc_t *)C->NodeList;

        /* build class nodelist */

        nindex = 0;

        /* FORMAT:  <HOST>[,<HOST>]... */

        ptr = MUStrTok(AVal,", \t\n",&TokPtr);

        while (ptr != NULL)
          {
          if (MNodeFind(ptr,&N) == FAILURE)
            {
            ptr = MUStrTok(NULL,", \t\n",&TokPtr);

            continue;
            }

          NL[nindex].N  = N;
          NL[nindex].TC = 1;

          ptr = MUStrTok(NULL,", \t\n",&TokPtr);
          }  /* END while (ptr != NULL) */
        }    /* END BLOCK */

        break;

      case mclaState:

        if (!strcmp(AVal,"true"))
          {
          C->State = mclsActive;

          C->IsDisabled = FALSE;
          }
        else
          {
          C->State = mclsStopped;

          C->IsDisabled = TRUE;
          }

        break;

      default:

        /* attr not supported */

        return(FAILURE);

        /*NOTREACHED*/

        break;
      }  /* END switch(aindex) */
    }

  return(SUCCESS);
  }  /* END __MSSSQueueSetAttr() */





int __MSSSJobSetAttr(

  mjob_t  *J,      /* I (modified) */
  int      VIndex, /* I */
  char    *AName,  /* I */
  char    *AVal,   /* I */
  mbool_t  IsUtilized, /* I */
  void    *E)      /* I (optional) */

  {
  int aindex;

  mreq_t *RQ;

  if ((J == NULL) || (AName == NULL))
    {
    return(FAILURE);
    }

  if ((aindex = MUGetIndex(AName,(const char **)MSSSJobAttr[VIndex],FALSE,0)) != 0)
    {
    if (AVal == NULL)
      {
      if (aindex != mjaAllocNodeList)
        {
        return(SUCCESS);
        }
      }

    switch(aindex)
      {
      case mjaJobName:
      case mjaJobID:
      case mjaUser:
      case mjaGroup:
      case mjaAccount:
      case mjaReqSMinTime:
      case mjaRMXString:
      case mjaStartTime:
      case mjaSubmitTime:
      case mjaCompletionTime:

        if (AVal == NULL)
          {
          return(SUCCESS);
          }

        MJobSetAttr(J,aindex,(void **)AVal,mdfString,mSet);

        break;

      case mjaReqProcs:

        if (IsUtilized == TRUE)
          {
          /* utilized procs NYI */

          aindex = mjaNONE;
          }

        MJobSetAttr(J,aindex,(void **)AVal,mdfString,mSet);

        break;

      case mjaReqNodes:

        if (IsUtilized == TRUE)
          {
          /* utilized nodes NYI */

          aindex = mjaNONE;
          }

        MJobSetAttr(J,aindex,(void **)AVal,mdfString,mSet);

        break;

      case mjaReqAWDuration:
      case mjaAWDuration:

        if (IsUtilized == TRUE)
          {
          aindex = mjaAWDuration;
          }
        else
          {
          aindex = mjaReqAWDuration;
          }

        MJobSetAttr(J,aindex,(void **)AVal,mdfString,mSet);

        break;

      case mjaAllocNodeList:

        {
        mxml_t *NE;

        char *ptr;

        char tmpBuf[MMAX_BUFFER];

        char *BPtr;
        int   BSpace;

        int   CTok;
 
        /* FORMAT:  <Node>X</Node>... */

        ptr = AVal;

        BPtr = tmpBuf;
        BSpace = sizeof(tmpBuf);

        BPtr[0] = '\0';

        /* convert XML to tasklist string */

        CTok = -1;

        while (MXMLGetChild((mxml_t *)E,"Node",&CTok,&NE) == SUCCESS)
          {
          if (tmpBuf[0] != '\0')
            MUStrNCat(&BPtr,&BSpace,",");

          MUStrNCat(&BPtr,&BSpace,NE->Val);
          }  /* END while (MXMLFromString() == SUCCESS) */

        MJobSetAttr(J,aindex,(void **)tmpBuf,mdfString,mSet);
        }  /* END BLOCK */

        break;

      case mjaState:

        switch(VIndex)
          {
          case msssV2_0:

            J->State = MUGetIndex((char *)AVal,MSSSJobState,FALSE,0);

            break;

          case msssV0_2:
          default:

            if ((J->State = MUGetIndex((char *)AVal,MJobState,FALSE,0)) == 0)
              {
              J->State = MUGetIndex((char *)AVal,MSSSJobState,FALSE,0);
              }

            break;
          }        /* END switch(VIndex) */

        break;

      default:

        /* attr not supported */

        return(FAILURE);

        /*NOTREACHED*/

        break;
      }  /* END switch(aindex) */
    }
  else if ((aindex = MUGetIndex(AName,(const char **)MSSSReqAttr[VIndex],FALSE,0)) != 0)
    {
    RQ = J->Req[0];  /* support single req SSS jobs only */

    switch(aindex)
      {
      case mrqaReqClass:

        MReqSetAttr(J,RQ,aindex,(void **)AVal,mdfString,mSet);

        break;

      case mrqaTCReqMin:

        if (IsUtilized == TRUE)
          {
          /* utilized procs NYI */

          aindex = mrqaNONE;
          }

        MReqSetAttr(J,RQ,aindex,(void **)AVal,mdfString,mSet);

        break;

      default:

        /* attr not supported */

        return(FAILURE);

        /*NOTREACHED*/

        break;
      }  /* END switch(aindex) */
    }    /* END else if (MUGetIndex() != 0) */
  else
    {
    /* check if contained w/in Requested/Utilized */

    mxml_t *AE;
    mxml_t *CE;

    mbool_t IsUtilized;

    int CTok;

    AE = (mxml_t *)E;

    if (!strcmp(AName,"Requested"))
      {
      IsUtilized = FALSE;
      }
    else if (!strcmp(AName,"Delivered"))
      {
      IsUtilized = TRUE;
      }
    else
      {
      /* unrecognized attribute */

      return(FAILURE);
      }

    CTok = -1;

    while (MXMLGetChild(AE,NULL,&CTok,&CE) == SUCCESS)
      {
      char UData[MMAX_NAME];
      char *AVal;

      AVal = CE->Val;

      /* check units */

      if (MXMLGetAttr(CE,"units",NULL,UData,sizeof(UData)) == SUCCESS)
        {
        char tmpData[MMAX_LINE];

        sprintf(tmpData,"%s%s",
          AVal,
          UData);

        /* adjust value by unit information */

        /* NYI */
        }

      __MSSSJobSetAttr(
        J,
        (J->SRM != NULL) ? J->SRM->Version : 0,
        CE->Name,
        CE->Val,
        IsUtilized,
        (void *)CE);
      }  /* END while (MXMLGetChild(AE,NULL,&CTok,&CE) == SUCCESS) */

    /* attribute not handled */

    return(FAILURE);
    }  /* END else ((aindex = MUGetIndex(AName)) != 0) */

  return(SUCCESS);
  }  /* END __MSSSJobSetAttr() */





int MSSSDoCommand(

  mpsi_t  *P,         /* I */
  char    *CmdString, /* I (optional) */
  char   **OBuf,      /* O (optional) */
  mxml_t **ODE)       /* O (optional) */

  {
  msocket_t *S;

  msocket_t  tmpS;

  char      *ServerHost = NULL;
  int        ServerPort = -1;

  enum MSocketProtocolEnum SocketProtocol = mspNONE;
  enum MWireProtocolEnum   WireProtocol   = mwpNONE;
  enum MChecksumAlgoEnum   CSAlgo         = mcsaNONE;

  long       Timeout;

  char      *CSKey = NULL;

  mxml_t   *SDE;

  const char *FName = "MSSSDoCommand";

  MDB(2,fSSS) MLog("%s(%s,%s,OBuf)\n",
    FName,
    (P != NULL) ? MSSSCName[P->Type] : "NULL",
    (CmdString != NULL) ? CmdString : "NULL");

  if (ODE != NULL)
    *ODE = NULL;

  if ((P == NULL) || 
      (P->Type <= mpstNONE))
    {
    MDB(2,fSSS) MLog("ALERT:    invalid peer type in %s\n",
      FName);

    return(FAILURE);
    }

  if ((OBuf != NULL) && (*OBuf != NULL))
    MUFree(OBuf);

  Timeout = (P->Timeout > 0) ? P->Timeout : 1;
  
  switch(P->Type)
    {
    case mpstWWW:

      memset(&tmpS,0,sizeof(tmpS));

      S          = &tmpS;

      ServerHost = P->HostName;

      ServerPort = (P->Port > 0) ? P->Port : 80;

      SocketProtocol = mspHTTPClient;

      CSKey      = NULL;

      break;

    case mpstQM:
    case mpstNM:
    case mpstSD:
    case mpstEM:
    case mpstAM:

      S          = P->S;

      ServerHost = P->HostName;
      ServerPort = P->Port;

      SocketProtocol = P->SocketProtocol;
      WireProtocol   = P->WireProtocol;

      CSKey      = P->CSKey;
      CSAlgo     = P->CSAlgo;

      break;

    default:

      /* invalid service requested */

      return(FAILURE);

      /*NOTREACHED*/

      break;
    }  /* END switch(P->Type) */

  /* free old data objects */

  if ((ServerHost == NULL) || (ServerHost[0] == '\0'))
    {
    /* remote access point not configured */

    MDB(2,fSSS) MLog("ALERT:    invalid server host in %s\n",
      FName);

    return(FAILURE);
    }

  if ((P->Type != mpstEM) && (P->Type != mpstWWW))
    {
    char tmpLine[MMAX_LINE];

    sprintf(tmpLine,"%s,%s;",
      MSSSCName[mpstSC],
      MSSSCName[P->Type]);

    MSysEMSubmit(
      &MSched.EM,
      (char *)MSSSCName[mpstSC],
      "comcom",
      tmpLine); 
    }

  if (S != NULL)
    {
    MSUFree(S);
    }

  if (S->sd <= 0)
    {
    /* obtain location/config info from discovery service */

    /* NYI */

    SDE = S->SDE;
 
    MSUInitialize(
      S,
      ServerHost,
      ServerPort,
      0,
      (1 << msftTCP));

    S->SDE = SDE;
 
    S->SocketProtocol = SocketProtocol;
    S->WireProtocol   = WireProtocol;

    if (CSKey != NULL)
      strcpy(S->CSKey,CSKey);

    if (CSAlgo != mcsaNONE)
      S->CSAlgo = CSAlgo;

    if (MSUConnect(S) == FAILURE)
      {
      char Msg[MMAX_LINE];

      MDB(0,fSSS) MLog("ERROR:    cannot connect to %s server '%s':%d\n",
        MSSSCName[P->Type],
        ServerHost,
        ServerPort);
 
      sprintf(Msg,"RMFAILURE:  cannot connect to %s server %s:%d (command: '%s')\n",
        MSSSCName[P->Type],
        ServerHost,
        ServerPort,
        (CmdString != NULL) ? CmdString : "<XML>");
 
      MSysRegEvent(Msg,0,0,1);

      return(FAILURE);
      }
    }    /* END if (S->sd <= 0) */

  if (P->Data != NULL)
    {
    S->URI = (char *)P->Data;
    }

  if (CmdString != NULL)
    {
    MSSSEncapsulateMessage(
      CmdString,
      P->Version,
      TRUE,
      &S->SBuffer,
      &S->SBufSize);
    }

  /* send data */

  if (MSUSendData(S,Timeout * 1000000,FALSE,FALSE) == FAILURE)
    {
    char Msg[MMAX_LINE];

    MSUDisconnect(S);

    MDB(0,fSSS) MLog("ERROR:    cannot send data to %s server '%s':%d\n",
      MSSSCName[P->Type],
      ServerHost,
      ServerPort);
 
    sprintf(Msg,"FAILURE:  cannot send data to %s server %s:%d (cmd: '%s')\n",
      MSSSCName[P->Type],
      ServerHost,
      ServerPort,
      (CmdString != NULL) ? CmdString : "<XML>");
 
    MSysRegEvent(Msg,0,0,1);

    return(FAILURE);
    }
  else
    {
    MDB(1,fSSS) MLog("INFO:     command sent to server\n");
 
    MDB(3,fSSS) MLog("INFO:     message sent: '%s'\n",
      (CmdString != NULL) ? CmdString : "<XML>");
    }  /* END else (MSUSendData(S,Timeout * 1000000,FALSE,FALSE) == FAILURE) */

  /* receive data */

  if (MSURecvData(S,Timeout * 1000000,FALSE) == FAILURE)
    {
    char Msg[MMAX_LINE];

    MSUDisconnect(S);

    if ((OBuf != NULL) && (S->SMsg != NULL))
      {
      MUStrDup(OBuf,S->SMsg);
      }
 
    MDB(0,fSSS) MLog("ERROR:    cannot receive response from %s server '%s':%d\n",
      MSSSCName[P->Type],
      ServerHost,
      ServerPort);
 
    sprintf(Msg,"FAILURE:  cannot receive response from %s server %s:%d (cmd: '%s')\n",
      MSSSCName[P->Type],
      ServerHost,
      ServerPort,
      (CmdString != NULL) ? CmdString : "<XML>");
 
    MSysRegEvent(Msg,0,0,1);

    return(FAILURE);
    }
  else
    {
    if ((P->Type != mpstEM) && (P->Type != mpstWWW))
      {
      char tmpLine[MMAX_LINE];

      sprintf(tmpLine,"%s,%s;",
        MSSSCName[P->Type],
        MSSSCName[mpstSC]);

      MSysEMSubmit(
        &MSched.EM,
        (char *)MSSSCName[mpstSC],
        "comcom",
        tmpLine);
      }

    MDB(1,fSSS) MLog("INFO:     response received from server\n");
 
    MDB(3,fSSS) MLog("INFO:     response received: '%s'\n",
      S->RBuffer);
    }  /* END else (MSURecvData(S,Timeout * 1000000,FALSE) == FAILURE) */

  if (OBuf != NULL)  
    *OBuf = strdup(S->RBuffer);

  if (ODE != NULL)
    MMOVEPTR((char *)S->RDE,(char *)*ODE);

  if ((S->SocketProtocol == mspHalfSocket) ||
      (S->SocketProtocol == mspSSSChallenge) ||
      (S->SocketProtocol == mspHTTP) ||
      (S->SocketProtocol == mspHTTPClient))
    {
    /* one time use socket */

    /* NOTE:  disconnecting socket frees xml/string buffer data */

    MSUDisconnect(S);
    }

  return(SUCCESS);
  }  /* END MSSSDoCommand() */




long MSSSGetResKVal(

  char *String)  /* I */

  {
  long  val; 
  char *tail;

  char  Mod;

  /* currently drop milliseconds */

  val = strtol(String,&tail,10);

  if (*tail == ':')   /* time resource -> convert to seconds */
    {
    char *tail2;

    val *= 3600;                               /* hours   */

    val += (strtol(tail + 1,&tail2,10) * 60);  /* minutes */

    if (*tail2 == ':')
      val += strtol(tail2 + 1,&tail,10);       /* seconds */

    return(val);
    }

  Mod = toupper(*tail);

  if (Mod == 'K')
    {
    return(val);
    }
  else if (Mod == 'M')
    {
    return(val << 10);
    }
  else if (Mod == 'G')
    {
    return(val << 20);
    }
  else if (Mod == 'T')
    {
    return(val << 30);
    }
#ifdef __M64
  else if (Mod == 'P')
    {
    return(val << 40);
    }
  else if (Mod == 'X')
    {
    return(val << 50);
    }
  else if (Mod == 'Z')
    {
    return(val << 60);
    }
#endif /* __M64 */
  else if (Mod == 'b')
    {
    return(val >> 10);
    }

  return(val);
  }  /* END MSSSGetResKVal() */




int MSSSQueueGetInfo(

  mnode_t *N,
  char     CClass[][MMAX_NAME],
  char     AClass[][MMAX_NAME])

  {
  int cindex;
  int CIndex;

  int ConsumedClasses;

  memset(N->ARes.PSlot,0,sizeof(N->ARes.PSlot));
  memset(N->CRes.PSlot,0,sizeof(N->CRes.PSlot));

  /* update configured classes */

  if (CClass[0][0] == '\0')
    {
    return(FAILURE);
    }

  for (cindex = 0;CClass[cindex][0] != '\0';cindex++)
    {
    MDB(8,fSSS) MLog("INFO:     configured class '%s' specified for node %s\n",
      CClass[cindex],
      N->Name);

    if ((CIndex = MUMAGetIndex(meClass,CClass[cindex],mAdd)) != FAILURE)
      {
      N->CRes.PSlot[CIndex].count++;
      N->CRes.PSlot[0].count++;
      }
    else
      {
      MDB(1,fSSS) MLog("ALERT:    cannot add configured class '%s' to node %s\n",
        CClass[cindex],
        N->Name);
      }
    }    /* END for (cindex) */

  /* update available classes */

  for (cindex = 0;AClass[cindex][0] != '\0';cindex++)
    {
    MDB(8,fSSS) MLog("INFO:     available class '%s' specified for node %s\n",
      AClass[cindex],
      N->Name);

    if ((CIndex = MUMAGetIndex(meClass,AClass[cindex],mAdd)) != FAILURE)
      {
      if (N->CRes.PSlot[CIndex].count > N->ARes.PSlot[CIndex].count)
        {
        N->ARes.PSlot[CIndex].count++;
        N->ARes.PSlot[0].count++;
        }
      else
        {
        MDB(1,fSSS) MLog("ALERT:    class '%s' available but not configured\n",
          AClass[cindex]);
        }
      }
    else
      {
      MDB(1,fSSS) MLog("ALERT:    cannot add available class '%s' to node %s\n",
        AClass[cindex],
        N->Name);
      }
    }    /* END for (cindex) */

  ConsumedClasses = 0;

  for (cindex = 1;cindex < MMAX_CLASS;cindex++)
    {
    ConsumedClasses += MAX(
      0,
      N->CRes.PSlot[cindex].count - N->ARes.PSlot[cindex].count);
    }

  if (((N->State == mnsIdle) &&
      (MUNumListGetCount(MMAX_PRIO_VAL,N->CRes.PSlot,N->ARes.PSlot,0,NULL) == FAILURE)) ||
      (MUNumListGetCount(MMAX_PRIO_VAL,N->ARes.PSlot,N->CRes.PSlot,0,NULL) == FAILURE))
    {
    /* SSS corruption */

    MDB(1,fSSS) MLog("ALERT:    %s node %s has class mismatch.  classes: %s\n",
      MAList[meNodeState][N->State],
      N->Name,
      MUNumListToString(N->ARes.PSlot,N->CRes.PSlot,NULL,NULL,0));

    if (MSched.Mode == msmNormal)
      {
      MOSSyslog(LOG_NOTICE,"node %s in state %s has class mismatch.  classes: %s\n",
        N->Name,
        MAList[meNodeState][N->State],
        MUNumListToString(N->ARes.PSlot,N->CRes.PSlot,NULL,NULL,0));
      }
    }

  /* adjust class by max procs */

  if (N->AP.HLimit[mptMaxProc][0] > 0)
    {
    N->CRes.PSlot[0].count =
      MIN(
        N->CRes.PSlot[0].count,
        N->AP.HLimit[mptMaxProc][0]);

    if ((N->CRes.PSlot[0].count - ConsumedClasses) > 0)
      {
      N->ARes.PSlot[0].count =
        MIN(
          N->ARes.PSlot[0].count,
          N->CRes.PSlot[0].count - ConsumedClasses);
      }
    }

  return(SUCCESS);
  }  /* END MSSSQueueGetInfo() */




int MSSSJobRequeue(

  mjob_t  *J,     /* I (modified) */
  mrm_t   *R,     /* I */
  mjob_t **JPeer, /* I (optional) */
  char    *Msg,   /* O */
  int     *SC)    /* O */

  {
  char *ErrMsg = NULL;

  mxml_t *E  = NULL;
  mxml_t *RE = NULL;
  mxml_t *CE = NULL;
  mxml_t *DE = NULL;
  mxml_t *JE = NULL;

  char *ptr;

  char *Response = NULL;

  char  CmdString[MMAX_BUFFER];

  char tmpJobName[MMAX_NAME];

  const char *FName = "MSSSJobRequeue";

  MDB(2,fSSS) MLog("%s(%s,%s,JPeer,Msg,SC)\n",
    FName,
    (J != NULL) ? J->Name : "NULL",
    (R != NULL) ? R->Name : "NULL");

  if ((J == NULL) || (R == NULL))
    {
    return(FAILURE);
    }
 
  if (MJOBISACTIVE(J) == FAILURE)
    {
    return(FAILURE);
    }

  MJobGetName(
    J,
    NULL,
    R,
    tmpJobName,
    sizeof(tmpJobName),
    mjnRMName);   

  switch(R->Version)
    {
    case msssV2_0:

      MXMLCreateE(&E,"Request");
      MXMLSetAttr(E,"action",(void *)MSSSJAName[R->Version][mrmJobRequeue],mdfString);
      MXMLSetAttr(E,"object",(void *)MSSSObjName[R->Version][mxoJob],mdfString);

      DE = NULL;
      MXMLCreateE(&DE,"Data");
      MXMLAddE(E,DE);

      JE = NULL;
      MXMLCreateE(&JE,MSSSObjName[R->Version][mxoJob]);
      MXMLAddE(DE,JE);

      CE = NULL;
      MXMLCreateE(&CE,MSSSJobAttr[R->Version][mjaJobID]);
      /* MXMLCreateE(&CE,"StepId"); */
      MXMLSetVal(CE,(void *)J->Name,mdfString);
      MXMLAddE(JE,CE);

      break;

    case msssV0_2:

      MXMLCreateE(&E,"Message");
      MXMLCreateE(&RE,"Request");

      MXMLAddE(E,RE);

      MXMLSetAttr(RE,"action",(void *)"requeue",mdfString);
      MXMLSetAttr(RE,"object",(void *)"job",mdfString);

      CE = NULL;
      MXMLCreateE(&CE,(char *)MSON[msonWhere]);
      MXMLSetAttr(CE,"name",(void *)MSSSJobAttr[R->Version][mjaJobID],mdfString);
      MXMLSetAttr(CE,"value",(void *)J->Name,mdfString);
      MXMLAddE(RE,CE);

      break;
  
    default:

      return(FAILURE);

      /*NOTREACHED*/

      break;
    }  /* END switch(R->Version) */

  MXMLToString(E,CmdString,sizeof(CmdString),NULL,TRUE);

  MXMLDestroyE(&E);

  if (MSSSDoCommand(
        &R->P[0],
        CmdString,
        &Response,
        NULL) == FAILURE)
    {
    /* job could not be requeued */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be requeued (command failed)\n",
      J->Name);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  if (((ptr = strstr(Response,"<Body")) == NULL) ||
       (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
    {
    /* cannot parse response */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be requeued (invalid response)\n",
      J->Name);

    MUFree(&Response);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  MUFree(&Response);

  if (MXMLGetChild(E,"Response",NULL,&RE) == FAILURE)
    {
    /* cannot parse response */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be requeued (cannot parse response)\n",
      J->Name);

    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;
    }

  /* check status code */

  if (MXMLGetChild(RE,"Status",NULL,&CE) == FAILURE)
    {
    /* cannot locate status code */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be requeued (cannot parse command status)\n",
      J->Name);

    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  if (strcmp(CE->Val,"true"))
    {
    /* command failed */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be requeued (request refused)\n",
      J->Name);

    if ((ErrMsg != NULL) && (MXMLGetChild(RE,"Message",NULL,&CE) == FAILURE))
      {
      strcpy(ErrMsg,CE->Val);
      }

    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  /* job successfully requeued */

  MSysEMSubmit(
    &MSched.EM,
    (char *)MSSSCName[mpstSC],
    (char *)"jobrequeue",
    (char *)J->Name);

  MDB(1,fSSS) MLog("INFO:     job '%s' successfully requeued\n",
    J->Name);

  return(SUCCESS);
  }  /* END MSSSJobRequeue() */




char *__MSSSGetErrMsg(

  int SD)  /* I */

  {
  static char *ptr;

  ptr = NULL;

  return(ptr);
  }  /* END __MSSSGetErrMsg() */




int MSSSJobSuspend(

  mjob_t *J,       /* I */
  mrm_t  *R,       /* I */
  char   *EMsg,    /* O (optional) */
  int    *SC)      /* O (optional) */

  {
  mxml_t *E  = NULL;
  mxml_t *RE = NULL;
  mxml_t *CE = NULL;
  mxml_t *DE = NULL;
  mxml_t *JE = NULL;

  char         Message[MMAX_LINE];

  char        *Response = NULL;

  char         CmdString[MMAX_BUFFER];

  char    *ptr;
  
  const char *FName = "MSSSJobSuspend";

  MDB(1,fSSS) MLog("%s(%s,%s,%s,SC)\n",
    FName,
    (J != NULL) ? J->Name : "NULL",
    (R != NULL) ? R->Name : "NULL",
    (EMsg != NULL) ? EMsg : "NULL");

  if ((J == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  if (MJOBISACTIVE(J) == FALSE)
    {
    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be suspended: (invalid state)\n",
      J->Name);

    return(FAILURE);
    }

  switch(R->Version)
    {
    case msssV2_0:
    default:

      MXMLCreateE(&E,MSON[msonRequest]);
      MXMLSetAttr(E,MSAN[msanAction],(void *)MSSSJAName[R->Version][mrmJobSuspend],mdfString);

      MSSSSetObject(E,MSSSObjName[R->Version][mxoJob],NULL);

      DE = NULL;
      MXMLCreateE(&DE,MSON[msonData]);
      MXMLAddE(E,DE);

      JE = NULL;
      MXMLCreateE(&JE,MSSSObjName[R->Version][mxoJob]);
      MXMLAddE(DE,JE);

      CE = NULL;
      MXMLCreateE(&CE,MSSSJobAttr[R->Version][mjaJobID]);
      /* MXMLCreateE(&CE,"StepId"); */
      MXMLSetVal(CE,(void *)J->Name,mdfString);
      MXMLAddE(JE,CE);

      break;
    }  /* END switch(R->Version) */

  MXMLToString(E,CmdString,sizeof(CmdString),NULL,TRUE);

  MXMLDestroyE(&E);

  if (MSSSDoCommand(
        &R->P[0],
        CmdString,
        &Response,
        NULL) == FAILURE)
    {
    /* job could not be suspended */

    R->FailIteration = MSched.Iteration;

    if (SC != NULL)
      *SC = mscRemoteFailure;

    return(FAILURE);
    }

  if (((ptr = strstr(Response,"<Body")) == NULL) ||
       (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
    {
    /* cannot parse response */

    MUFree(&Response);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  MUFree(&Response);

  if (MXMLGetChild(E,MSON[msonResponse],NULL,&RE) == FAILURE)
    {
    /* cannot parse response */

    MXMLDestroyE(&E);

    R->FailIteration = MSched.Iteration;
    }

  /* check status code */

  if (MXMLGetChild(RE,"Status",NULL,&CE) == FAILURE)
    {
    /* cannot locate status code */

    MXMLDestroyE(&E);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  if ((CE->Val != NULL) && (strcmp(CE->Val,"true")))
    {
    mxml_t *tE = NULL;

    /* command failed */

    char tmpMsg[MMAX_LINE];
    int  tmpSC;

    if (MXMLGetChild(CE,"Message",NULL,&tE) == SUCCESS)
      {
      MUStrCpy(tmpMsg,tE->Val,sizeof(tmpMsg));
      }
    else
      {
      strcpy(tmpMsg,"NoMessage");
      }

    if (MXMLGetChild(RE,"Code",NULL,&tE) == SUCCESS)
      {
      tmpSC = (int)strtol(tE->Val,NULL,0);
      }
    else
      {
      tmpSC = -1;
      }

    snprintf(Message,MMAX_LINE,"cannot suspend job - RM failure, rc: %d, msg: '%s'",
      tmpSC,
      tmpMsg);

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be suspended: RM failure (%s)\n",
      J->Name,
      Message);

    if (EMsg != NULL)
      strcpy(EMsg,Message);

    MUStrDup(&J->Message,Message);

    MXMLDestroyE(&E);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }  /* END if (strcmp(C->Val,"true")) */

  /* job successfully started */

  MXMLDestroyE(&E);

  MSysEMSubmit(
    &MSched.EM,
    (char *)MSSSCName[mpstSC],
    (char *)"jobsuspend",
    (char *)J->Name);

  /* adjust state */

  MJobSetState(J,mjsSuspended);

  MDB(1,fSSS) MLog("INFO:     job '%s' successfully suspended\n",
    J->Name);

  return(SUCCESS);
  }  /* END MSSSJobSuspend() */




int MSSSJobResume(

  mjob_t *J,       /* I */
  mrm_t  *R,       /* I */
  char   *EMsg,    /* O (optional) */
  int    *SC)      /* O (optional) */

  {
  char         Message[MMAX_LINE];

  char        *Response = NULL;

  char         CmdString[MMAX_BUFFER];

  mxml_t *E  = NULL;
  mxml_t *RE = NULL;
  mxml_t *CE = NULL;
  mxml_t *DE = NULL;
  mxml_t *JE = NULL;

  char *ptr;

  const char *FName = "MSSSJobResume";

  MDB(1,fSSS) MLog("%s(%s,%s,%s,SC)\n",
    FName,
    (J != NULL) ? J->Name : "NULL",
    (R != NULL) ? R->Name : "NULL",
    (EMsg != NULL) ? EMsg : "NULL");

  if ((J == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  if (J->State != mjsSuspended)
    {
    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be resumed: (invalid state)\n",
      J->Name);

    return(FAILURE);
    }

  switch(R->Version)
    {
    case msssV2_0:

      MXMLCreateE(&E,MSON[msonRequest]);
      MXMLSetAttr(E,MSAN[msanAction],(void *)MSSSJAName[R->Version][mrmJobResume],mdfString);
      MSSSSetObject(E,MSSSObjName[R->Version][mxoJob],NULL);

      DE = NULL;
      MXMLCreateE(&DE,MSON[msonData]);
      MXMLAddE(E,DE);

      JE = NULL;
      MXMLCreateE(&JE,MSSSObjName[R->Version][mxoJob]);
      MXMLAddE(DE,JE);

      CE = NULL;
      MXMLCreateE(&CE,MSSSJobAttr[R->Version][mjaJobID]);
      /* MXMLCreateE(&CE,"StepId"); */
      MXMLSetVal(CE,(void *)J->Name,mdfString);
      MXMLAddE(JE,CE);

      break;

    case msssV0_2:
    default:

      MXMLCreateE(&E,"Message");
      MXMLCreateE(&RE,"Request");

      MXMLAddE(E,RE);

      MXMLSetAttr(RE,"action",(void *)"resume",mdfString);
      MXMLSetAttr(RE,"object",(void *)"job",mdfString);

      CE = NULL;
      MXMLCreateE(&CE,(char *)MSON[msonWhere]);
      MXMLSetAttr(CE,"name",(void *)MSSSJobAttr[R->Version][mjaJobID],mdfString);
      MXMLSetAttr(CE,"value",(void *)J->Name,mdfString);
      MXMLAddE(RE,CE);

      break;
    }  /* END switch(R->Version) */

  MXMLToString(E,CmdString,sizeof(CmdString),NULL,TRUE);

  MXMLDestroyE(&E);

  if (MSSSDoCommand(
        &R->P[0],
        CmdString,
        &Response,
        NULL) == FAILURE)
    {
    /* job could not be suspended */

    R->FailIteration = MSched.Iteration;

    if (SC != NULL)
      *SC = mscRemoteFailure;

    return(FAILURE);
    }

  if (((ptr = strstr(Response,"<Body")) == NULL) ||
       (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
    {
    /* cannot parse response */

    MUFree(&Response);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  MUFree(&Response);

  if (MXMLGetChild(E,MSON[msonResponse],NULL,&RE) == FAILURE)
    {
    /* cannot parse response */

    MXMLDestroyE(&E);

    R->FailIteration = MSched.Iteration;
    }

  /* check status code */

  if (MXMLGetChild(RE,"Status",NULL,&CE) == FAILURE)
    {
    /* cannot locate status code */

    MXMLDestroyE(&E);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  if ((CE->Val != NULL) && (strcmp(CE->Val,"true")))
    {
    /* command failed */

    mxml_t *tE = NULL;

    char tmpMsg[MMAX_LINE];
    int  tmpSC;

    if (MXMLGetChild(CE,"Message",NULL,&tE) == SUCCESS)
      {
      MUStrCpy(tmpMsg,tE->Val,sizeof(tmpMsg));
      }
    else
      {
      strcpy(tmpMsg,"NoMessage");
      }

    if (MXMLGetChild(CE,"Code",NULL,&tE) == SUCCESS)
      {
      tmpSC = (int)strtol(tE->Val,NULL,0);
      }
    else
      {
      tmpSC = -1;
      }

    snprintf(Message,MMAX_LINE,"cannot resume job - RM failure, rc: %d, msg: '%s'",
      tmpSC,
      tmpMsg);

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be resumed: RM failure (%s)\n",
      J->Name,
      Message);

    if (EMsg != NULL)
      strcpy(EMsg,Message);

    MUStrDup(&J->Message,Message);

    MXMLDestroyE(&E);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }  /* END if (strcmp(C->Val,"true")) */

  /* job successfully resumed */

  MXMLDestroyE(&E);

  MSysEMSubmit(
    &MSched.EM,
    (char *)MSSSCName[mpstSC],
    (char *)"jobresume",
    (char *)J->Name);

  /* adjust state */

  MJobSetState(J,mjsSuspended);

  MDB(1,fSSS) MLog("INFO:     job '%s' successfully resume\n",
    J->Name);

  return(SUCCESS);
  }  /* END MSSSJobResume() */





int MSSSJobCheckpoint(

  mjob_t  *J,        /* I */
  mrm_t   *R,        /* I */
  mbool_t  TerminateJobs,  /* I */
  char    *ErrMsg,   /* O (optional) */
  int     *SC)       /* O (optional) */

  {
  mxml_t *E  = NULL;
  mxml_t *RE = NULL;
  mxml_t *CE = NULL;
  mxml_t *DE = NULL;
  mxml_t *JE = NULL;

  char *ptr;

  char *Response = NULL;

  char  CmdString[MMAX_BUFFER];

  const char *FName = "MSSSJobCheckpoint";

  MDB(1,fSSS) MLog("%s(%s,%s,%s,ErrMsg,SC)\n",
    FName,
    (J != NULL) ? J->Name : "NULL",
    (R != NULL) ? R->Name : "NULL",
    (TerminateJobs == TRUE) ? "TRUE" : "FALSE");

  if ((J == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  switch(R->Version)
    {
    case msssV2_0:

      MXMLCreateE(&E,MSON[msonRequest]);
      MXMLSetAttr(E,MSAN[msanAction],(void *)MSSSJAName[R->Version][mrmJobCheckpoint],mdfString);
      MSSSSetObject(E,MSSSObjName[R->Version][mxoJob],NULL);

      DE = NULL;
      MXMLCreateE(&DE,MSON[msonData]);
      MXMLAddE(E,DE);

      JE = NULL;
      MXMLCreateE(&JE,MSSSObjName[R->Version][mxoJob]);
      MXMLAddE(DE,JE);

      CE = NULL;
      MXMLCreateE(&CE,MSSSJobAttr[R->Version][mjaJobID]);
      /* MXMLCreateE(&CE,"StepId"); */
      MXMLSetVal(CE,(void *)J->Name,mdfString);
      MXMLAddE(JE,CE);

      break;

    case msssV0_2:

      MXMLCreateE(&E,"Message");
      MXMLCreateE(&RE,"Request");

      MXMLAddE(E,RE);

      MXMLSetAttr(RE,"action",(void *)"checkpoint",mdfString);
      MXMLSetAttr(RE,"object",(void *)"job",mdfString);

      CE = NULL;
      MXMLCreateE(&CE,(char *)MSON[msonWhere]);
      MXMLSetAttr(CE,"name",(void *)MSSSJobAttr[R->Version][mjaJobID],mdfString);
      MXMLSetAttr(CE,"value",(void *)J->Name,mdfString);
      MXMLAddE(RE,CE);

      break;
    }  /* END switch(R->Version) */

  MXMLToString(E,CmdString,sizeof(CmdString),NULL,TRUE);

  MXMLDestroyE(&E);

  if (MSSSDoCommand(
        &R->P[0],
        CmdString,
        &Response,
        NULL) == FAILURE)
    {
    /* job could not be cancelled */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be checkpointed (command failed)\n",
      J->Name);

    R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  if (((ptr = strstr(Response,"<Body")) == NULL) ||
       (MXMLFromString(&E,ptr,NULL,NULL) == FAILURE))
    {
    /* cannot parse response */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be checkpointed (invalid response)\n",
      J->Name);

    MUFree(&Response);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  MUFree(&Response);

  if (MXMLGetChild(E,"Response",NULL,&RE) == FAILURE)
    {
    /* cannot parse response */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be checkpointed (cannot parse response)\n",
      J->Name);

    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;
    }

  /* check status code */

  if (MXMLGetChild(RE,"Status",NULL,&CE) == FAILURE)
    {
    /* cannot locate status code */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be checkpointed (cannot parse command status)\n",
      J->Name);

    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  if (strcmp(CE->Val,"true"))
    {
    /* command failed */

    MDB(0,fSSS) MLog("ERROR:    job '%s' cannot be checkpointed (request refused)\n",
      J->Name);

    if ((ErrMsg != NULL) && (MXMLGetChild(RE,"Message",NULL,&CE) == FAILURE))
      {
      strcpy(ErrMsg,CE->Val);
      }

    MXMLDestroyE(&E);

    if (R->FailIteration == -1)
      R->FailIteration = MSched.Iteration;

    return(FAILURE);
    }

  /* job successfully checkpointed */

  MSysEMSubmit(
    &MSched.EM,
    (char *)MSSSCName[mpstSC],
    (char *)"jobcheckpoint",
    (char *)J->Name);

  MDB(1,fSSS) MLog("INFO:     job '%s' successfully checkpointed\n",
    J->Name);

  return(SUCCESS);
  }  /* END MSSSJobCheckpoint() */





#ifdef MNOT

int __MSSSGetTaskList(

  char   TaskString[],
  short  TaskList[],
  int   *TaskCount,
  int   *TPN)

  {
  int     index;
  int     tindex;

  char   *ptr;
  char   *tptr;

  char   *TokPtr;

  mnode_t *N;
  char    tmpHostName[MMAX_NAME];

  int     ppn;

  long    tmpL;
  char   *tail;

  /* FORMAT:  <HOSTNAME>[:ppn=<X>][<HOSTNAME>[:ppn=<X>]]... */

  MDB(5,fSSS) MLog("__MSSSGetTaskList(%s,%s,%s,%s)\n",
    (TaskString != NULL) ? TaskString : "NULL",
    (TaskList != NULL) ? "TaskList" : "NULL",
    (TaskCount != NULL) ? "TaskCount" : "NULL",
    (TPN != NULL) ? "TPN" : "NULL");

  if ((TaskString == NULL) || (TaskList == NULL))
    {
    return(FAILURE);
    }
 
  ptr = MUStrTok(TaskString,"+ \t",&TokPtr);

  tindex = 0;

  /* NOTE:  only handles one 'ppn' distribution per job */

  ppn = 1;

  while (ptr != NULL)
    {
    ppn = 1;

    strcpy(tmpHostName,ptr);

    if ((tptr = strchr(tmpHostName,'/')) != NULL)
      *tptr = '\0';

    if ((tptr = strchr(tmpHostName,':')) != NULL)
      {
      *tptr = '\0';

      tptr++;

      if (strncmp(tptr,"ppn=",strlen("ppn=")) == 0)
        {
        tptr += strlen("ppn=");

        ppn = strtol(tptr,NULL,10);
        }
      }

    tmpL = strtol(tmpHostName,&tail,0);

    if ((tmpL != 0) && (*tail == '\0'))
      {
      /* nodecount specified */
      }
    else
      { 
      if (MNodeFind(tmpHostName,&N) == FAILURE)
        {
        MDB(1,fSSS) MLog("ALERT:    cannot locate host '%s' for job hostlist\n",
          tmpHostName);
        }
      else
        {
        for (index = 0;index < ppn;index++)
          {
          TaskList[tindex] = N->Index;

          tindex++;
          }
        }
      }

    ptr = MUStrTok(NULL,"+ \t",&TokPtr);
    }    /* END while (ptr != NULL) */

  TaskList[tindex] = -1;

  if (TaskCount != NULL)
    *TaskCount = tindex;

  if (TPN != NULL)
    *TPN = ppn;

  MDB(7,fSSS) MLog("INFO:     %d task(s) located for job\n",
    tindex);

  if (tindex == 0)
    {
    return(FAILURE);
    }

  return(SUCCESS);
  }    /* END __MSSSGetTaskList() */




int MSSSJobModify(
 
  mjob_t *J,         /* I (modified) */
  char   *Name,      /* I */
  char   *Resource,  /* I */
  char   *Value)     /* I */
 
  {
  int rc;
 
  char  tmpJName[MMAX_NAME];
  char *ErrMsg;

  if (J == NULL)
    {
    return(FAILURE);
    }
 
  MJobGetName(
    J,
    NULL,
    (J->SRM != NULL) ? J->SRM : &MRM[0],
    tmpJName,
    sizeof(tmpJName),
    mjnRMName);

  /* NYI */
 
  rc = -1;
 
  if (rc != 0)
    {
    ErrMsg = NULL;
 
    MDB(0,fSSS) MLog("ERROR:    cannot set job '%s' attr '%s:%s' to '%s' (rc: %d '%s')\n",
      J->Name,
      Name,
      Resource,
      Value,
      rc,
      ErrMsg);
 
    return(FAILURE);
    }
 
  return(SUCCESS);
  }  /* END MSSSJobModify() */

#endif /* MNOT */




int MSSSGetNodeState() { return(SUCCESS); }
int MSSSNodeFree() { return(SUCCESS); }


int MSSSProcessQueueInfo() 

  { 
  return(SUCCESS); 
  }  /* END MSSSProcessQueueInfo() */




int MSSSEncapsulateMessage(

  char     *DataString, /* I */
  char     *Version,    /* I */
  mbool_t   IsRequest,  /* I */
  char    **SBuf,       /* O */
  long     *SBufSize)   /* I */

  {
  if (SBuf != NULL)
    *SBuf = NULL;

  if ((DataString == NULL) ||
      (SBuf == NULL) ||
      (SBufSize == NULL))
    {
    return(FAILURE);
    }

  if (Version == NULL)
    {
    *SBuf     = DataString;
    *SBufSize = strlen(DataString) + 1;

    return(SUCCESS);
    }

  if (!strcmp(Version,"SSS2.0"))
    {
    /* NYI */
    }

  return(SUCCESS);
  }  /* END MSSSEncapsulateMessage() */




int MSSSExtractMessage(

  char     *DataString, /* I */
  char     *Version,    /* I */
  mbool_t  *IsRequest,
  char    **SBuf,       /* O */
  long     *SBufSize)   /* I */

  {
  if (SBuf != NULL)
    *SBuf = NULL;

  if ((DataString == NULL) ||
      (SBuf == NULL) ||
      (SBufSize == NULL))
    {
    return(FAILURE);
    }

  if (Version == NULL)
    {
    /* FORMAT:  <Message><Response/></Message> */

    *SBuf     = DataString;
    *SBufSize = strlen(DataString) + 1;

    return(SUCCESS);
    }

  if (!strcmp(Version,"SSS2.0"))
    {
    /* FORMAT:  <Envelope><Request/></Envelope> */

    /* NYI */
    }

  return(SUCCESS);
  }  /* END MSSSExtractMessage() */




int MSSSJobSubmit(

  char                 *Data,   /* I (job description) */
  mrm_t                *R,      /* I */
  mjob_t              **JP,     /* O (optional) */
  char                 *JobID,  /* O (optional) */
  char                 *EMsg,   /* O (optional) */
  enum MStatusCodeEnum *SC)     /* O (optional) */

  {
  mxml_t *JE;
  mxml_t *DE;

  sssrm_t *S;

  mjob_t   *J;

  short TaskList[MMAX_TASK_PER_JOB + 1];

  char     *SJobID = "sss-submit";

  const char *FName = "MSSSJobSubmit";

  MDB(1,fSSS) MLog("%s(%s,%s,%s,%s,SC)\n",
    FName,
    (Data != NULL) ? "Data" : "NULL",
    (R != NULL) ? R->Name : "NULL",
    (JP != NULL) ? "JP" : "NULL",
    (EMsg != NULL) ? "EMsg" : "NULL");

  if ((Data == NULL) || (R == NULL))
    {
    return(FAILURE);
    }

  if (!MISSET(R->Flags,mrmfLocalQueue))
    {
    /* can only submit to local queue */

    return(FAILURE);
    }

  S = (sssrm_t *)R->S;

  if (S->WorkloadBuffer == NULL)
    {
    /* allocate local queue workload buffer */

    DE = NULL;

    MXMLCreateE(&DE,(char *)MSON[msonData]);

    S->WorkloadBuffer = (char *)DE;
    }
  else
    {
    DE = (mxml_t *)S->WorkloadBuffer;
    }

  /* parse job */

  if (MXMLFromString(&JE,Data,FALSE,0) == FAILURE)
    {
    /* cannot process job */

    return(FAILURE);
    }

  /* validate job */

  if (MJobCreate(SJobID,TRUE,&J) == FAILURE)
    {
    MDB(1,fSSS) MLog("ERROR:    job buffer is full  (ignoring job '%s')\n",
      SJobID);

    return(FAILURE);
    }

  /* if new job, load data */

  MRMJobPreLoad(J,SJobID,R->Index);

  MJobSetAttr(J,mjaRMJID,(void **)SJobID,mdfString,mSet);

  if (MSSSJobLoad(SJobID,JE,&J,TaskList,R) == FAILURE)
    {
    MDB(1,fSSS) MLog("ALERT:    cannot load SSS job '%s'\n",
      SJobID);

    /* cannot validate job */

    return(FAILURE);
    }

  MRMJobPostLoad(J,TaskList,R);

  /* add job to internal queue */

  MXMLAddE(DE,JE);

  /* decision to process job immediately or wait for next iteration */
  /* is made at higher level */

  return(SUCCESS);
  }  /* END MSSSJobSubmit() */

  



int MSSSAddWhere(

  mxml_t *E,     /* I (modified) */
  char   *Name,  /* I */
  char   *Value) /* I */

  {
  mxml_t *WE;

  if ((E == NULL) || (Name == NULL) || (Value == NULL))
    {
    return(FAILURE);
    }

  WE = NULL;

  MXMLCreateE(&WE,MSON[msonWhere]);
  MXMLSetAttr(WE,MSAN[msanName],(void *)Name,mdfString);

  if (Value != NULL)
    {
#ifdef __MSSS30
    MXMLSetVal(WE,(void *)Value,mdfString);
#else /* __MSSS30 */
    MXMLSetAttr(WE,MSAN[msanValue],(void *)Value,mdfString);
#endif /* __MSSS30 */
    }  /* END if (Value != NULL) */

  MXMLAddE(E,WE);

  return(SUCCESS);
  }  /* END MSSSAddWhere() */





int MSSSGetWhere(

  mxml_t  *E,     /* I */
  mxml_t **WEP,   /* O (optional) */
  int     *STok,  /* I/O (optional) */
  char    *Name,  /* O */
  int      NSize, /* I */
  char    *Value, /* O */
  int      VSize) /* I */

  {
  mxml_t *WE;

  if (WEP != NULL)
    *WEP = NULL;

  if (Name != NULL)
    Name[0] = '\0';

  if (Value != NULL)
    Value[0] = '\0';

  if (E == NULL)
    {
    return(FAILURE);
    }

  if (MXMLGetChild(E,MSON[msonWhere],STok,&WE) == FAILURE)
    {
    return(FAILURE);
    }

  MXMLGetAttr(WE,MSAN[msanName],NULL,Name,NSize);

#ifdef __MSSS30

  if (WE->Val != NULL)
    MUStrCpy(Value,WE->Val,VSize);

#else /* __MSSS30 */

  MXMLGetAttr(WE,MSAN[msanValue],NULL,Value,VSize);

#endif /* __MSSS30 */

  if (WEP != NULL)
    *WEP = WE;

  return(SUCCESS);
  }  /* END MSSSGetWhere() */





int MSSSAddSet(

  mxml_t  *E,     /* I (modified) */
  char    *Name,  /* I */
  char    *Value, /* I */
  mxml_t **SEP)   /* O (optional) */

  {
  mxml_t *SE;

  if (SEP != NULL)
    *SEP = NULL;

  if ((E == NULL) || (Name == NULL))
    {
    return(FAILURE);
    }

  SE = NULL;

  MXMLCreateE(&SE,MSON[msonSet]);
  MXMLSetAttr(SE,MSAN[msanName],(void *)Name,mdfString);

  if (Value != NULL)
    {
#ifdef __MSSS30
    MXMLSetVal(SE,(void *)Value,mdfString);
#else /* __MSSS30 */
    MXMLSetAttr(SE,MSAN[msanValue],(void *)Value,mdfString);
#endif /* __MSSS30 */
    }  /* END if (Value != NULL) */

  MXMLAddE(E,SE);

  if (SEP != NULL)
    {
    *SEP = SE;
    }

  return(SUCCESS);
  }  /* END MSSSAddSet() */





int MSSSAddGet(

  mxml_t  *E,     /* I (modified) */
  char    *Name,  /* I */
  mxml_t **GEP)   /* O (optional) */

  {
  mxml_t *GE;

  if (GEP != NULL)
    *GEP = NULL;

  if ((E == NULL) || (Name == NULL))
    {
    return(FAILURE);
    }

  GE = NULL;

  MXMLCreateE(&GE,MSON[msonGet]);

  if (Name != NULL)
    {
#ifdef __MSSS30
    MXMLSetAttr(GE,MSAN[msanName],(void *)Name,mdfString);
#else /* __MSSS30 */
    MXMLSetAttr(GE,MSAN[msanName],(void *)Name,mdfString);
#endif /* __MSSS30 */
    }  /* END if (Value != NULL) */

  MXMLAddE(E,GE);

  if (GEP != NULL)
    {
    *GEP = GE;
    }

  return(SUCCESS);
  }  /* END MSSSAddGet() */




int MSSSGetSet(

  mxml_t  *E,     /* I */
  mxml_t **SEP,   /* O (optional) */
  int     *STok,  /* I/O (optional) */
  char    *Name,  /* O */
  int      NSize, /* I */
  char    *Value, /* O */
  int      VSize) /* I */

  {
  mxml_t *SE;

  if (SEP != NULL)
    *SEP = NULL;

  if (Name != NULL)
    Name[0] = '\0';

  if (Value != NULL)
    Value[0] = '\0';

  if (E == NULL)
    {
    return(FAILURE);
    }

  if (MXMLGetChild(E,MSON[msonSet],STok,&SE) == FAILURE)
    {
    return(FAILURE);
    }

  MXMLGetAttr(SE,MSAN[msanName],NULL,Name,NSize);

#ifdef __MSSS30

  if (SE->Val != NULL)
    MUStrCpy(Value,SE->Val,VSize);

#else /* __MSSS30 */

  MXMLGetAttr(SE,MSAN[msanValue],NULL,Value,VSize);

#endif /* __MSSS30 */

  if (SEP != NULL)
    *SEP = SE;

  return(SUCCESS);
  }  /* END MSSSGetSet() */





int MSSSGetGet(

  mxml_t  *E,     /* I */
  mxml_t **GEP,   /* O (optional) */
  int     *GTok,  /* I/O (optional) */
  char    *Name,  /* O */
  int      NSize) /* I */

  {
  mxml_t *GE;

  if (GEP != NULL)
    *GEP = NULL;

  if (Name != NULL)
    Name[0] = '\0';

  if (E == NULL)
    {
    return(FAILURE);
    }

  if (MXMLGetChild(E,MSON[msonGet],GTok,&GE) == FAILURE)
    {
    return(FAILURE);
    }

#ifdef __MSSS30

  if (GE->Val != NULL)
    MUStrCpy(Name,GE->Val,NSize);

#else /* __MSSS30 */

  MXMLGetAttr(GE,MSAN[msanName],NULL,Name,NSize);

#endif /* __MSSS30 */

  if (GEP != NULL)
    *GEP = GE;

  return(SUCCESS);
  }  /* END MSSSGetGet() */






int MSSSSetObject(

  mxml_t *E,      /* I */
  char   *OName,  /* I */
  char   *Value)  /* I (optional) */

  {
#ifndef MOPT

  if ((E == NULL) || (OName == NULL))
    {
    return(FAILURE);
    }

#endif /* MOPT */

#ifdef __MSSS30

  {
  mxml_t *OE = NULL;

  MXMLCreateE(&OE,MSON[msonObject]);
  MXMLSetVal(OE,(void *)OName,mdfString);

  MXMLAddE(E,OE);
  }  /* END BLOCK */

#else /* __MSSS30 */

  MXMLSetAttr(E,MSAN[msanObject],(void *)OName,mdfString);

#endif /* __MSSS30 */

  return(SUCCESS);
  }  /* END MSSSSetObject() */




int MSSSGetObject(

  mxml_t *E,     /* I */
  char   *OName) /* O (min size MMAX_NAME) */

  {
  if ((E == NULL) || (OName == NULL))
    {
    return(FAILURE);
    }

  OName[0] = '\0';

#ifdef __MSSS30

  {
  mxml_t *OE;

  if (MXMLGetChild(E,(char *)MSON[msonObject],NULL,&OE) == SUCCESS)
    {
    if (OE->Val != NULL)
      {
      MUStrCpy(OName,OE->Val,MMAX_NAME);

      return(SUCCESS);
      }
    }
  }  /* END BLOCK */

#else /* __MSSS30 */

  if (MXMLGetAttr(E,MSAN[msanObject],NULL,OName,MMAX_NAME) == SUCCESS)
    {
    return(SUCCESS);
    }

#endif /* __MSSS30 */

  return(FAILURE);
  }  /* END MSSSGetObject() */



int MSSSCheckStatus(

  mxml_t    *E,    /* I */
  enum MSFC *SC,   /* O (optional) */
  char      *EMsg) /* O (optional) */

  {
  mxml_t *SE;
  mxml_t *CE;

#ifndef MOPT

  if (E == NULL)
    {
    if (SC != NULL)
      *SC = msfEGMisc;

    if (EMsg != NULL)
      strcpy(EMsg,"internal error");

    return(FAILURE);
    }

#endif /* MOPT */

  /* set defaults */

  if (SC != NULL)
    *SC = msfENone;

  if (EMsg != NULL)
    EMsg[0] = '\0';

#ifdef __MSSS30

  /*
     <Status>
       <Value>Success|Warning|Failure</Value>  -- required
       <Code>X</Code>                          -- required
       <Message>X</Message>                    -- optional
     </Status>
  */

  if (MXMLGetChild(E,"Status",NULL,&SE) == FAILURE)
    {
    /* cannot locate status element */

    MDB(3,fSSS) MLog("ERROR:    cannot locate message status element\n");

    if (SC != NULL)
      *SC = msfEGMessage;

    if (EMsg != NULL)
      strcpy(EMsg,"no status");

    return(FAILURE);
    }

  if ((MXMLGetChild(SE,"Message",NULL,&CE) == FAILURE) ||
      (CE->Val == NULL))
    {
    /* cannot locate status message */

    MDB(6,fSSS) MLog("INFO:   cannot locate status message element\n");

    return(FAILURE);
    }
  else
    {
    if (EMsg != NULL)
      MUStrCpy(EMsg,CE->Val,MMAX_LINE);
    }

  if ((MXMLGetChild(SE,"Code",NULL,&CE) == FAILURE) ||
      (CE->Val == NULL))
    {
    /* cannot locate status code */

    MDB(3,fSSS) MLog("WARNING:  cannot locate status code element\n");

    if (SC != NULL)
      *SC = msfEGMessage;

    if (EMsg != NULL)
      MUStrCpy(EMsg,"no status code",MMAX_LINE);

    return(FAILURE);
    }

  if (SC != NULL)
    {
    *SC = strtol(CE->Val,NULL,0);
    }

  if (MXMLGetChild(SE,"Value",NULL,&CE) == FAILURE)
    {
    /* cannot locate status code */

    MDB(3,fSSS) MLog("WARNING:  cannot locate status value element\n");

    if (EMsg != NULL)
      MUStrCpy(EMsg,"no status value",MMAX_LINE);

    return(FAILURE);
    }

  if ((CE->Val == NULL) || !strcmp(CE->Val,"Failure"))
    {
    /* command failed */

    MDB(3,fSSS) MLog("WARNING:  request failed\n");

    return(FAILURE);
    }

#else /* __MSSS30 */

  if (MXMLGetChild(E,"Status",NULL,&CE) == FAILURE)
    {
    /* cannot locate status code */

    MDB(3,fSSS) MLog("ERROR:    cannot locate message status element\n");

    if (SC != NULL)
      *SC = msfEGMessage;

    if (EMsg != NULL)
      strcpy(EMsg,"no status");

    return(FAILURE);
    }

  if (SC != NULL)
    {
    mxml_t *tE;

    if ((MXMLGetChild(CE,"Code",NULL,&tE) == SUCCESS) &&
        (tE->Val != NULL))
      {
      *SC = strtol(tE->Val,NULL,0);
      }
    else
      {
      *SC = msfEGMessage;
      }
    }

  if ((CE->Val == NULL) || strcmp(CE->Val,"true"))
    {
    /* command failed */

    MDB(3,fSSS) MLog("ERROR:    request refused\n");

    if ((EMsg != NULL) && (MXMLGetChild(E,"Message",NULL,&CE) == FAILURE))
      {
      strcpy(EMsg,CE->Val);
      }

    return(FAILURE);
    }

#endif /* !__MSSS30 */

  return(SUCCESS);
  }  /* END MSSSCheckStatus() */



/* END MSSSI.c */


